# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import FastTextLangId
from nemo_curator.utils.distributed_utils import get_client, read_data
from nemo_curator.utils.file_utils import (
    get_all_files_paths_under,
    separate_by_metadata,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def load_dataset(input_data_dir):
    files = list(get_all_files_paths_under(input_data_dir, keep_extensions="jsonl"))
    raw_data = read_data(files, file_type="jsonl", backend="pandas", add_filename=True)
    dataset = DocumentDataset(raw_data)

    return dataset


def main(args):
    # Params
    multilingual_data_path = "/path/to/multilingual"
    language_separated_output_path = "/path/to/lang_separated"
    cleaned_data_output_path = "/path/to/cleaned"

    # Download a fastText language identification model
    # and see a list of supported languages here:
    # https://fasttext.cc/docs/en/language-identification.html
    model_path = "/path/to/model.bin"
    language_field = "language"

    # Prepare samples for the classifier
    client = get_client(**ArgumentHelper.parse_client_args(args))

    # Filter data
    multilingual_dataset = load_dataset(multilingual_data_path)
    language_id_pipeline = nc.ScoreFilter(
        FastTextLangId(model_path), score_field=language_field, score_type="object"
    )
    filtered_dataset = language_id_pipeline(multilingual_dataset)

    # Remove the language score
    filtered_dataset.df[language_field] = filtered_dataset.df[language_field].apply(
        lambda score: score[1], meta=(None, str)
    )

    # Split the dataset by language
    language_stats = separate_by_metadata(
        filtered_dataset.df,
        language_separated_output_path,
        metadata_field=language_field,
    ).compute()


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    return ArgumentHelper(parser).add_distributed_args()


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, Iterator, List, Optional, Tuple, Union

import spacy
from presidio_analyzer.nlp_engine import (
    NerModelConfiguration,
    NlpArtifacts,
    SpacyNlpEngine,
)
from spacy import Language

logger = logging.getLogger("presidio-analyzer")


class CustomNlpEngine(SpacyNlpEngine):

    def __init__(
        self,
        models: Optional[List[Dict[str, str]]] = None,
        ner_model_configuration: Optional[NerModelConfiguration] = None,
    ):
        super().__init__(models, ner_model_configuration)
        self.nlp: Dict[str, Language] = None

    def load(self) -> None:
        """Load the spaCy NLP model."""
        logger.debug(f"Loading SpaCy models: {self.models}")

        self.nlp = {}
        # Download spaCy model if missing
        for model in self.models:
            self._validate_model_params(model)
            self._download_spacy_model_if_needed(model["model_name"])
            self.nlp[model["lang_code"]] = spacy.load(
                model["model_name"], enable=["ner"]
            )

    def process_batch(
        self,
        texts: Union[List[str], List[Tuple[str, object]]],
        language: str,
        as_tuples: bool = False,
        batch_size: int = 32,
    ) -> Iterator[Optional[NlpArtifacts]]:
        """Execute the NLP pipeline on a batch of texts using spacy pipe.

        :param texts: A list of texts to process.
        :param language: The language of the texts.
        :param as_tuples: If set to True, inputs should be a sequence of
            (text, context) tuples. Output will then be a sequence of
            (doc, context) tuples. Defaults to False.
        :param batch_size: The batch size.
        """

        if not self.nlp:
            raise ValueError("NLP engine is not loaded. Consider calling .load()")

        texts = [str(text) for text in texts]
        docs = self.nlp[language].pipe(
            texts, as_tuples=as_tuples, batch_size=batch_size
        )
        for doc in docs:
            yield doc.text, self._doc_to_nlp_artifact(doc, language)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

from nemo_curator.modules.fuzzy_dedup._mapbuckets import _MapBuckets
from nemo_curator.utils.distributed_utils import get_client, get_num_workers
from nemo_curator.utils.fuzzy_dedup_utils.io_utils import (
    get_bucket_ddf_from_parquet_path,
    get_text_ddf_from_json_path_with_blocksize,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def get_anchor_and_output_map_info(
    input_data_paths,
    input_bucket_path,
    text_ddf_blocksize,
    num_files,
    num_workers,
    shuffle_type,
    input_bucket_field,
    input_id_field,
    input_text_field,
    input_meta,
):
    """
    Get anchor docs with bucket info
    Args:
        input_data_paths: list of paths to input data
        input_bucket_path: path to input buckets
        text_ddf_blocksize: blocksize for text ddf
        num_files: number of files to read
        num_workers: number of workers
        shuffle_type: type of shuffle to use
    Returns:
        ddf_anchor_docs_with_bk
    """
    ddf_text = get_text_ddf_from_json_path_with_blocksize(
        input_data_paths=input_data_paths,
        num_files=num_files,
        blocksize=text_ddf_blocksize,
        id_column=input_id_field,
        text_column=input_text_field,
        input_meta=input_meta,
    )
    ddf_bk = get_bucket_ddf_from_parquet_path(
        input_bucket_path=input_bucket_path, num_workers=num_workers
    )
    map_buckets = _MapBuckets(
        id_fields=["dataset_id", "doc_id"],
        bucket_field=input_bucket_field,
        text_field=input_text_field,
    )
    ddf_anchor_docs_with_bk = map_buckets.map_buckets_with_anchors(
        documents_df=ddf_text, buckets_df=ddf_bk, shuffle_type=shuffle_type
    )
    return ddf_anchor_docs_with_bk


def attach_args():
    description = """
    Takes the buckets generated from minhashes and uses
    document length information to create a coarse mapping of mapping multiple
    buckets to a logical partition by using a modified bin packing algorithm.
    """
    parser = argparse.ArgumentParser(
        description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.parse_gpu_dedup_args()

    argumentHelper.add_arg_input_meta()
    argumentHelper.add_arg_output_dir()
    argumentHelper.add_arg_text_ddf_blocksize()
    parser.add_argument(
        "--input-bucket-dir",
        type=str,
        help="The directory containing bucket information files.",
    )
    parser.add_argument(
        "--input-bucket-field",
        type=str,
        default="_bucket_id",
        help="Name of the column containing minhashes.",
    )
    parser.add_argument(
        "--shuffle-type",
        type=str,
        default="tasks",
        help="Type of shuffle to use before writing to Parquet.",
    )

    return parser


def jaccard_get_output_map_workflow(
    client,
    input_data_paths,
    input_bucket_path,
    output_anchor_docs_with_bk_path,
    text_ddf_blocksize,
    num_files,
    shuffle_type,
    input_bucket_field,
    input_id_field,
    input_text_field,
    input_meta,
):
    """
    Workflow for jaccard shuffle
    Args:
        client: dask client
        input_data_paths: list of paths to input data
        input_bucket_path: path to input buckets
        output_anchor_docs_with_bk_path: path to save anchor docs with bucket info
        text_ddf_blocksize: blocksize for text ddf
        num_files: number of files to read
        parts_per_worker: number of parts per worker
        shuffle_type: type of shuffle to use before writing to parquet
    """
    num_workers = get_num_workers(client)
    ddf_anchor_docs_with_bk = get_anchor_and_output_map_info(
        input_data_paths,
        input_bucket_path,
        text_ddf_blocksize,
        num_files,
        num_workers,
        shuffle_type,
        input_bucket_field,
        input_id_field,
        input_text_field,
        input_meta=input_meta,
    )
    ddf_anchor_docs_with_bk.to_parquet(
        output_anchor_docs_with_bk_path,
        write_index=False,
    )


def main(args):
    input_data_paths = args.input_data_dirs
    input_bucket_path = args.input_bucket_dir
    OUTPUT_PATH = args.output_dir
    output_anchor_docs_with_bk_path = os.path.join(
        OUTPUT_PATH, "anchor_docs_with_bk.parquet"
    )
    client = get_client(**ArgumentHelper.parse_client_args(args))
    print(f"Num Workers = {get_num_workers(client)}", flush=True)
    print("Connected to dask cluster", flush=True)
    print("Running jaccard map buckets script", flush=True)
    print(f"Args = {args}")
    st = time.time()
    jaccard_get_output_map_workflow(
        client,
        input_data_paths,
        input_bucket_path,
        output_anchor_docs_with_bk_path,
        args.text_ddf_blocksize,
        args.num_files,
        args.shuffle_type,
        args.input_bucket_field,
        args.input_json_id_field,
        args.input_json_text_field,
        args.input_meta,
    )
    et = time.time()
    print(f"Bucket Mapping time taken = {et-st} s")


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Literal


class DocumentModifier(ABC):
    def __init__(self):
        super().__init__()
        self._name = self.__class__.__name__
        self._sentences = None
        self._paragraphs = None
        self._ngrams = None

    @abstractmethod
    def modify_document(self, text):
        pass

    @property
    def backend(self) -> Literal["pandas", "cudf", "any"]:
        """
        The dataframe backend the modifier operates on.
        Can be 'pandas', 'cudf', or 'any'. Defaults to 'pandas'.
        Returns:
            str: A string representing the dataframe backend the modifier needs as input
        """
        return "pandas"



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from nemo_curator.modifiers import DocumentModifier

THREE_OR_MORE_NEWLINES_REGEX = re.compile(r"(\n){3,}")
THREE_OR_MORE_WINDOWS_NEWLINES_REGEX = re.compile(r"(\r\n){3,}")


class NewlineNormalizer(DocumentModifier):
    """
    Replaces 3 or more consecutive newline characters with only 2 newline characters.
    """

    def __init__(self):
        super().__init__()

    def modify_document(self, text):
        text = THREE_OR_MORE_NEWLINES_REGEX.sub("\n\n", text)
        text = THREE_OR_MORE_WINDOWS_NEWLINES_REGEX.sub("\r\n\r\n", text)
        return text



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

from nemo_curator.services.conversation_formatter import ConversationFormatter


class NoFormat(ConversationFormatter):

    def format_conversation(self, conv: List[dict]) -> str:
        if len(conv) != 1:
            raise ValueError(
                "There must be exactly one turn in the conversation to use NoFormat"
            )

        turn = conv[0]

        if turn["role"] != "user":
            raise ValueError(
                "Conversation turn 0 is not 'user'. All even number turns should be."
            )

        return turn["content"]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from nemo_curator.download import download_common_crawl
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    # Params
    start_snapshot = "2021-04"
    end_snapshot = "2021-10"
    output_directory = "/path/to/output"

    # Only download 10 shards as an example
    url_limit = 10

    # Set up Dask client
    client = get_client(**ArgumentHelper.parse_client_args(args))

    # Download and sample data
    common_crawl = download_common_crawl(
        output_directory, start_snapshot, end_snapshot, url_limit=url_limit
    )
    sample = common_crawl.df.sample(frac=10 / len(common_crawl))

    # Inspect the samples
    print(sample.compute())


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    return ArgumentHelper(parser).add_distributed_args()


if __name__ == "__main__":
    main(attach_args().parse_args())



import gzip
import json
import os

import cudf
import dask.bag as db


def convert_single_file(input_output_paths):
    input_path, output_path = input_output_paths

    with gzip.open(input_path, "rt", encoding="utf-8") as f_in:
        with open(output_path, "w", encoding="utf-8") as f_out:
            for line in f_in:
                try:
                    # Parse each line as a separate JSON object
                    item = json.loads(line)
                    # Write the JSON object to the .jsonl file
                    json.dump(item, f_out)
                    f_out.write("\n")
                except json.JSONDecodeError as e:
                    print(f"Error decoding JSON in file {input_path}: {e}")
                    continue


def convert_json_gz_to_jsonl(input_dir, output_dir, partition_size=2):
    # Ensure the output directory exists
    os.makedirs(output_dir, exist_ok=True)

    # List all .json.gz files in the input directory
    file_paths = []
    for filename in os.listdir(input_dir):
        if filename.endswith(".json.gz"):
            input_path = os.path.join(input_dir, filename)
            output_filename = (
                os.path.splitext(os.path.splitext(filename)[0])[0] + ".jsonl"
            )
            output_path = os.path.join(output_dir, output_filename)
            file_paths.append((input_path, output_path))

    # Create a Dask bag from the file paths and apply the function in parallel
    bag = db.from_sequence(file_paths, partition_size=partition_size)
    bag.map(convert_single_file).compute()


def convert_str_id_to_int(df, id_column="id"):
    """
    Converts the legacy id format "dataset_name-0000034"
    type of ID into 2 int based ID's
    """
    dx = df[id_column].str.rsplit("-", n=1, expand=True)
    df["doc_id"] = dx[1].astype("int64").values
    df["dataset_id"] = dx[0].hash_values()
    return df


def get_dataframe_complement(original_df, filtered_df):
    def partition_complement(part_original_df, partition_info=None):
        if not partition_info:
            return part_original_df
        part_filtered_df = filtered_df.get_partition(partition_info["number"])
        complement_mask = ~part_original_df.index.isin(part_filtered_df.index.persist())
        complement_df = part_original_df[complement_mask]
        return complement_df

    return original_df.map_partitions(partition_complement)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Literal

from nemo_curator.datasets.doc_dataset import DocumentDataset
from nemo_curator.modules.base import BaseModule


class ToBackend(BaseModule):
    """
    A module for moving dataframes between backends.
    """

    def __init__(self, backend: Literal["pandas", "cudf"]) -> None:
        """
        Constructs a ToBackend module

        Args:
            backend (str): The backend to transfer the dataset to. Can be "pandas" or "cudf"
        """
        super().__init__(input_backend="any")
        if backend not in BaseModule.SUPPORTED_BACKENDS:
            raise ValueError(
                f"Backend must be either 'pandas' or 'cudf', but got {backend}"
            )
        self.backend = backend

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        return DocumentDataset(dataset.df.to_backend(self.backend))



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import warnings
from typing import Any, Coroutine, Dict, List, Optional

import pandas as pd
from openai import AsyncOpenAI
from tqdm.asyncio import tqdm

from nemo_curator.modifiers import DocumentModifier
from nemo_curator.utils.decorators import batched
from nemo_curator.utils.distributed_utils import load_object_on_worker
from nemo_curator.utils.llm_pii_utils import (
    JSON_SCHEMA,
    PII_LABELS,
    get_system_prompt,
    redact,
    validate_entity,
)

__all__ = ["AsyncLLMPiiModifier"]


class AsyncLLMInference:
    """A class for redacting PII via asynchronous LLM inference"""

    def __init__(
        self,
        base_url: str,
        api_key: Optional[str],
        model: str,
        system_prompt: str,
    ):
        self.client = AsyncOpenAI(base_url=base_url, api_key=api_key)
        self.model = model
        self.system_prompt = system_prompt

    async def infer(self, text: str) -> List[Dict[str, str]]:
        """Invoke LLM to get PII entities"""

        text = text.strip()
        messages = [
            {"role": "system", "content": self.system_prompt},
            {"role": "user", "content": text},
        ]

        response = await self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            # The field guided_json is unsupported at the root level
            # and must be included in the nvext object field
            extra_body={"nvext": {"guided_json": JSON_SCHEMA}},
            stream=False,
            max_tokens=4096,
        )

        assistant_message = response.choices[0].message.content

        # Parse results
        try:
            entities = json.loads(assistant_message)
            if not entities:
                # LLM returned valid JSON but no entities discovered
                return []
            else:
                # Check that each entity returned is valid
                return [e for e in entities if validate_entity(e, text)]
        except json.decoder.JSONDecodeError:
            return []


class AsyncLLMPiiModifier(DocumentModifier):
    """
    This class is the entry point to using the LLM-based PII de-identification module.
    It works with the `Modify` functionality as shown below:

    dataframe = pd.DataFrame({"text": ["Sarah and Ryan went out to play", "Jensen is the CEO of NVIDIA"]})
    dd = dask.dataframe.from_pandas(dataframe, npartitions=1)
    dataset = DocumentDataset(dd)

    modifier = AsyncLLMPiiModifier(
        # Endpoint for the user's NIM
        base_url="http://0.0.0.0:8000/v1",
        api_key="API KEY (if needed)",
        model="meta/llama-3.1-70b-instruct",
        # The user may engineer a custom prompt if desired
        system_prompt=None,
        pii_labels=PII_LABELS,
        language="en",
        max_concurrent_requests=10,
    )

    modify = Modify(modifier)
    modified_dataset = modify(dataset)
    modified_dataset.df.to_json("output_files/*.jsonl", lines=True, orient="records")

    """

    def __init__(
        self,
        base_url: str,
        api_key: Optional[str] = None,
        model: str = "meta/llama-3.1-70b-instruct",
        system_prompt: Optional[str] = None,
        pii_labels: Optional[List[str]] = None,
        language: str = "en",
        max_concurrent_requests: Optional[int] = None,
    ):
        """
        Initialize the AsyncLLMPiiModifier

        Args:
            base_url (str): The base URL for the user's NIM
            api_key (Optional[str]): The API key for the user's NIM, if needed.
                Default is None.
            model (str): The model to use for the LLM.
                Default is "meta/llama-3.1-70b-instruct".
            system_prompt (Optional[str]): The system prompt to feed into the LLM.
                If None, a default system prompt is used.
                Default prompt has been fine-tuned for "meta/llama-3.1-70b-instruct".
            pii_labels (Optional[List[str]]): The PII labels to identify and remove from the text.
                See documentation for full list of PII labels.
                Default is None, which means all PII labels will be used.
            language (str): The language to use for the LLM.
                Default is "en" for English. If non-English, it is recommended
                to provide a custom system prompt.
            max_concurrent_requests (Optional[int]): The maximum number of concurrent requests to make to the LLM.
                Default is None, which means no limit.

        """
        super().__init__()

        self.base_url = base_url
        self.api_key = api_key
        self.model = model
        self.max_concurrent_requests = max_concurrent_requests

        if system_prompt is not None and pii_labels is not None:
            warnings.warn(
                "Custom system_prompt and custom pii_labels were both provided, "
                "but the PII labels should already be included in the system prompt. "
                "The pii_labels will be ignored."
            )

        if pii_labels is None:
            pii_labels = PII_LABELS
        if system_prompt is None:
            self.system_prompt = get_system_prompt(pii_labels)
        else:
            self.system_prompt = system_prompt

        if language != "en" and system_prompt is None:
            warnings.warn(
                "The default system prompt is only available for English text. "
                "For other languages, please provide a custom system prompt. "
                "Please refer to the default system prompt as a guide: "
                "\n"
                f"{get_system_prompt(pii_labels)}"
                "\n"
                "In particular, please ensure that the JSON schema is included in the system prompt exactly as shown: "
                "\n"
                f"{str(JSON_SCHEMA)}"
            )
        if language == "en" and system_prompt is not None:
            warnings.warn(
                "Using the default system prompt is strongly recommended for English text. "
                "If you are customizing the system prompt, please refer to the default system prompt as a guide: "
                f"{get_system_prompt(pii_labels)}"
                "\n"
                "In particular, please ensure that the JSON schema is included in the system prompt exactly as shown: "
                "\n"
                f"{str(JSON_SCHEMA)}"
            )

    @batched
    def modify_document(self, text: pd.Series):
        self._inferer_key = f"inferer_{id(self)}"
        inferer = load_object_on_worker(self._inferer_key, self.load_inferer, {})
        pii_entities_lists = asyncio.run(self.call_inferer(text, inferer))
        text_redacted = self.batch_redact(text, pii_entities_lists)
        return text_redacted

    def load_inferer(self):
        """Helper function to load the asynchronous LLM"""
        inferer: AsyncLLMInference = AsyncLLMInference(
            base_url=self.base_url,
            api_key=self.api_key,
            model=self.model,
            system_prompt=self.system_prompt,
        )

        return inferer

    async def call_inferer(self, text: pd.Series, inferer: AsyncLLMInference):
        tasks = [inferer.infer(prompt) for prompt in text]
        pii_entities_lists = await self._gather(tasks)
        return pii_entities_lists

    async def _gather(
        self, requests: List[Coroutine[Any, Any, List[str]]]
    ) -> List[str]:
        max_requests = self.max_concurrent_requests
        if max_requests is None:
            max_requests = len(requests)

        final_list = []
        for i in tqdm(range(0, len(requests), max_requests)):
            request_slice = requests[i : i + max_requests]
            result = await tqdm.gather(*request_slice)
            final_list.extend(result)

        return final_list

    def batch_redact(
        self, text: pd.Series, pii_entities_lists: List[List[Dict[str, str]]]
    ):
        redacted_texts = [
            redact(text_str, pii_entities)
            for text_str, pii_entities in zip(text, pii_entities_lists)
        ]
        return pd.Series(redacted_texts, index=text.index)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import time

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modules import ExactDuplicates
from nemo_curator.utils.distributed_utils import get_client, write_to_disk
from nemo_curator.utils.script_utils import ArgumentHelper


def pre_imports():
    import cudf  # noqa: F401


def main(args):

    dataset_dir = "/path/to/data"
    log_dir = "./"
    output_dir = "./"
    dataset_id_field = "id"
    dataset_text_field = "text"
    client = get_client(**ArgumentHelper.parse_client_args(args))
    backend = "cudf" if args.device == "gpu" else "pandas"

    if args.device == "gpu":
        client.run(pre_imports)

    t0 = time.time()
    input_dataset = DocumentDataset.read_json(
        dataset_dir, backend=backend, blocksize="1GiB", files_per_partition=None
    )

    exact_dup = ExactDuplicates(
        logger=log_dir,
        id_field=dataset_id_field,
        text_field=dataset_text_field,
        # Decides whether output of the module is deduplicated dataset or duplicates
        # If true, you should set cache_dir for performance improvement
        perform_removal=False,
        # cache_dir=output_dir  # Optionally write the output to disk
    )

    # When perform_removal=False, it will only call .identify_duplicates() and return the list of duplicate IDs.
    # When perform_removal=True, then exact_dup outputs the dataset with the duplicates removed.
    # It will behave by calling .identify_duplicates() and .remove() in sequence.
    duplicates = exact_dup(
        dataset=input_dataset
    )  # or exact_dup.identify_duplicates(input_dataset)

    # If caching, result is a path to the output dataset.
    if isinstance(duplicates, str):
        duplicates = DocumentDataset.read_parquet(duplicates, backend=backend)

    result = exact_dup.remove(input_dataset, duplicates)
    write_to_disk(result, output_dir, output_type="parquet")
    print(time.time() - t0)


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    return ArgumentHelper(parser).add_distributed_args()


if __name__ == "__main__":
    main(attach_args().parse_args())



import dask.dataframe as dd
import pandas as pd
from dask.distributed import Client, LocalCluster

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset


def list_to_dataset(documents, col_name="text", npartitions=2):
    data = {col_name: documents}
    pdf = pd.DataFrame(data)

    return DocumentDataset(dd.from_pandas(pdf, npartitions=npartitions))


def all_equal(left_dataset, right_dataset):
    left_result = left_dataset.df.compute()
    right_result = right_dataset.df.compute()

    l_cols = set(left_result.columns)
    r_cols = set(right_result.columns)
    assert l_cols == r_cols
    for col in left_result.columns:
        left = left_result[col].reset_index(drop=True)
        right = right_result[col].reset_index(drop=True)
        assert all(left == right), f"Mismatch in {col} column.\n{left}\n{right}\n"


class TestShuffleNondeterministic:
    def test_shuffle(self):
        # Single threaded Dask is the only way to guarantee shuffle determinism
        # Docs: https://docs.dask.org/en/latest/generated/dask.dataframe.DataFrame.shuffle.html
        with LocalCluster(n_workers=1, threads_per_worker=1) as cluster:
            with Client(cluster):
                original_dataset = list_to_dataset(
                    ["one", "two", "three", "four", "five"]
                )
                expected_dataset = list_to_dataset(
                    ["two", "five", "three", "one", "four"]
                )
                shuffle = nc.Shuffle(seed=42)
                result_dataset = shuffle.shuffle_nondeterministic(original_dataset)
                all_equal(expected_dataset, result_dataset)

    def test_new_partitions(self):
        with LocalCluster(n_workers=1, threads_per_worker=1) as cluster:
            with Client(cluster):
                original_dataset = list_to_dataset(
                    ["one", "two", "three", "four", "five"], npartitions=3
                )
                expected_dataset = list_to_dataset(
                    ["two", "five", "three", "one", "four"], npartitions=3
                )
                shuffle = nc.Shuffle(seed=42, npartitions=2)
                result_dataset = shuffle.shuffle_nondeterministic(original_dataset)
                all_equal(expected_dataset, result_dataset)

    def test_filename(self):
        with LocalCluster(n_workers=1, threads_per_worker=1) as cluster:
            with Client(cluster):
                original_dataset = list_to_dataset(
                    ["one", "two", "three", "four", "five"], npartitions=1
                )
                original_dataset.df["file_name"] = "original.jsonl"

                expected_data = {
                    "text": ["one", "two", "three", "five", "four"],
                    "file_name": [
                        "file_0000000000.jsonl",
                        "file_0000000000.jsonl",
                        "file_0000000000.jsonl",
                        "file_0000000001.jsonl",
                        "file_0000000001.jsonl",
                    ],
                }
                pdf = pd.DataFrame(expected_data)
                expected_dataset = DocumentDataset(dd.from_pandas(pdf, npartitions=2))

                shuffle = nc.Shuffle(seed=42, npartitions=2)
                result_dataset = shuffle.shuffle_nondeterministic(original_dataset)
                all_equal(expected_dataset, result_dataset)

    def test_custom_filenames(self):
        with LocalCluster(n_workers=1, threads_per_worker=1) as cluster:
            with Client(cluster):
                original_dataset = list_to_dataset(
                    ["one", "two", "three", "four", "five"], npartitions=1
                )
                original_dataset.df["file_name"] = "original.jsonl"

                expected_data = {
                    "text": ["one", "two", "three", "five", "four"],
                    "file_name": [
                        "my_0.test",
                        "my_0.test",
                        "my_0.test",
                        "my_1.test",
                        "my_1.test",
                    ],
                }
                pdf = pd.DataFrame(expected_data)
                expected_dataset = DocumentDataset(dd.from_pandas(pdf, npartitions=2))

                def filename_fn(x):
                    return f"my_{x}.test"

                shuffle = nc.Shuffle(
                    seed=42, npartitions=2, partition_to_filename=filename_fn
                )
                result_dataset = shuffle.shuffle_nondeterministic(original_dataset)
                all_equal(expected_dataset, result_dataset)

    def test_shuffle_no_seed(self):
        original_dataset = list_to_dataset(["one", "two", "three", "four", "five"])
        shuffle = nc.Shuffle()
        result_dataset = shuffle(original_dataset)
        assert len(result_dataset.df.compute()) == 5


class TestShuffleDeterministic:
    def test_shuffle(self):
        original_dataset = list_to_dataset(["one", "two", "three", "four", "five"])
        expected_dataset = list_to_dataset(["five", "four", "three", "one", "two"])
        shuffle = nc.Shuffle(seed=42)
        result_dataset = shuffle(original_dataset)
        all_equal(expected_dataset, result_dataset)

    def test_new_partitions(self):
        original_dataset = list_to_dataset(
            ["one", "two", "three", "four", "five"], npartitions=3
        )
        expected_dataset = list_to_dataset(
            ["four", "three", "five", "one", "two"], npartitions=3
        )
        shuffle = nc.Shuffle(seed=42, npartitions=2)
        result_dataset = shuffle(original_dataset)
        all_equal(expected_dataset, result_dataset)

    def test_filename(self):
        original_dataset = list_to_dataset(
            ["one", "two", "three", "four", "five"], npartitions=1
        )
        original_dataset.df["file_name"] = "original.jsonl"

        expected_data = {
            "text": ["four", "five", "three", "one", "two"],
            "file_name": [
                "file_0000000000.jsonl",
                "file_0000000001.jsonl",
                "file_0000000001.jsonl",
                "file_0000000001.jsonl",
                "file_0000000001.jsonl",
            ],
        }
        pdf = pd.DataFrame(expected_data)
        expected_dataset = DocumentDataset(dd.from_pandas(pdf, npartitions=2))

        shuffle = nc.Shuffle(seed=42, npartitions=2)
        result_dataset = shuffle(original_dataset)
        all_equal(expected_dataset, result_dataset)

    def test_custom_filenames(self):
        original_dataset = list_to_dataset(
            ["one", "two", "three", "four", "five"], npartitions=1
        )
        original_dataset.df["file_name"] = "original.jsonl"

        expected_data = {
            "text": ["four", "five", "three", "one", "two"],
            "file_name": [
                "my_0.test",
                "my_1.test",
                "my_1.test",
                "my_1.test",
                "my_1.test",
            ],
        }
        pdf = pd.DataFrame(expected_data)
        expected_dataset = DocumentDataset(dd.from_pandas(pdf, npartitions=2))

        def filename_fn(x):
            return f"my_{x}.test"

        shuffle = nc.Shuffle(seed=42, npartitions=2, partition_to_filename=filename_fn)
        result_dataset = shuffle(original_dataset)
        all_equal(expected_dataset, result_dataset)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from nemo_curator.modifiers import DocumentModifier

MARKDOWN_BOLD_REGEX = r"\*\*(.*?)\*\*"
MARKDOWN_ITALIC_REGEX = r"\*(.*?)\*"
MARKDOWN_UNDERLINE_REGEX = r"_(.*?)_"
MARKDOWN_LINK_REGEX = r"\[.*?\]\((.*?)\)"


class MarkdownRemover(DocumentModifier):
    """
    Removes Markdown formatting in a document including bold, italic, underline, and URL text.
    """

    def __init__(self):
        super().__init__()

    def modify_document(self, text: str) -> str:
        lines = text.split("\n")
        new_lines = []
        for line in lines:
            line = re.sub(MARKDOWN_BOLD_REGEX, r"\1", line)  # **text**
            line = re.sub(MARKDOWN_ITALIC_REGEX, r"\1", line)  # *text*
            line = re.sub(MARKDOWN_UNDERLINE_REGEX, r"\1", line)  # _text_
            line = re.sub(MARKDOWN_LINK_REGEX, r"\1", line)  # [text](url)
            new_lines.append(line)

        return "\n".join(new_lines)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .bitext_filter import BitextFilter
from .classifier_filter import (
    FastTextLangId,
    FastTextQualityFilter,
    QualityEstimationFilter,
)
from .code import (
    AlphaFilter,
    GeneralCommentToCodeFilter,
    HTMLBoilerplateFilter,
    NumberOfLinesOfCodeFilter,
    PerExtensionFilter,
    PythonCommentToCodeFilter,
    TokenizerFertilityFilter,
    XMLHeaderFilter,
)
from .doc_filter import DocumentFilter, import_filter
from .heuristic_filter import (
    BoilerPlateStringFilter,
    BulletsFilter,
    CommonEnglishWordsFilter,
    EllipsisFilter,
    HistogramFilter,
    LengthRatioFilter,
    LongWordFilter,
    MeanWordLengthFilter,
    NonAlphaNumericFilter,
    NumbersFilter,
    ParenthesesFilter,
    PornographicUrlsFilter,
    PunctuationFilter,
    RepeatedLinesByCharFilter,
    RepeatedLinesFilter,
    RepeatedParagraphsByCharFilter,
    RepeatedParagraphsFilter,
    RepeatingDuplicateNGramsFilter,
    RepeatingTopNGramsFilter,
    SubstringFilter,
    SymbolsToWordsFilter,
    TokenCountFilter,
    UrlsFilter,
    WhiteSpaceFilter,
    WordCountFilter,
    WordsWithoutAlphabetsFilter,
)
from .synthetic import AnswerabilityFilter, EasinessFilter

__all__ = [
    "BitextFilter",
    "DocumentFilter",
    "import_filter",
    "FastTextLangId",
    "FastTextQualityFilter",
    "NonAlphaNumericFilter",
    "SymbolsToWordsFilter",
    "NumbersFilter",
    "UrlsFilter",
    "BulletsFilter",
    "WhiteSpaceFilter",
    "ParenthesesFilter",
    "LongWordFilter",
    "WordCountFilter",
    "BoilerPlateStringFilter",
    "MeanWordLengthFilter",
    "RepeatedLinesFilter",
    "RepeatedParagraphsFilter",
    "RepeatedLinesByCharFilter",
    "RepeatedParagraphsByCharFilter",
    "RepeatingTopNGramsFilter",
    "RepeatingDuplicateNGramsFilter",
    "PunctuationFilter",
    "EllipsisFilter",
    "CommonEnglishWordsFilter",
    "WordsWithoutAlphabetsFilter",
    "PornographicUrlsFilter",
    "PythonCommentToCodeFilter",
    "GeneralCommentToCodeFilter",
    "NumberOfLinesOfCodeFilter",
    "TokenizerFertilityFilter",
    "XMLHeaderFilter",
    "AlphaFilter",
    "HTMLBoilerplateFilter",
    "PerExtensionFilter",
    "LengthRatioFilter",
    "HistogramFilter",
    "QualityEstimationFilter",
    "AnswerabilityFilter",
    "EasinessFilter",
    "TokenCountFilter",
    "SubstringFilter",
]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import shutil
import time
from typing import Optional, Union

import cudf
import cupy as cp
import dask.dataframe as dd
import numpy as np
from cuml.dask.cluster import KMeans

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir
from nemo_curator.utils.semdedup_utils import (
    COSINE_DIST_TO_CENT_COL,
    L2_DIST_TO_CENT_COL,
    add_l2_cosine_dist_to_centroid,
    get_array_from_df,
    normalize_embeddings_col_in_df,
)


# Clustering module
class ClusteringModel:
    def __init__(
        self,
        id_column: str = "id",
        max_iter: int = 100,
        n_clusters: int = 1000,
        clustering_output_dir: str = "./clustering_results",
        embedding_column: str = "embeddings",
        random_state: int = 1234,
        clustering_input_partition_size: Optional[str] = "2gb",
        logger: Union[logging.Logger, str] = "./",
        profile_dir: Optional[str] = None,
        keep_all_columns: bool = False,
    ):
        """
        Initializes the ClusteringModel with the provided settings for semantic clustering to help semantic deduplication.

        Args:
            id_column (str): Column name used as the identifier in the dataset.
                Default is "id".
            max_iter (int): Maximum iterations for clustering. The more iterations, the better the clustering.
                Default is 100.
            n_clusters (int): Number of clusters. Default is 1000.
            clustering_output_dir (str): Location to save clustering results.
                Default is "./clustering_results".
            embedding_column (str): The column name that stores the embeddings.
                Default is "embeddings".
            random_state (int): KMeans random state used for reproducibility.
                Default is 1234.
            clustering_input_partition_size (Optional[str]): The size of data partition with which to run KMeans.
                Default is "2gb". If None, then the dataset is not repartitioned.
            logger (Union[logging.Logger, str]): Existing logger to log to, or a path to a log directory.
                Default is "./".
            profile_dir (Optional[str]): If specified, directory to write Dask profile.
                Default is None.

        """
        self.id_col = id_column
        self.max_iter = max_iter
        self.n_clusters = n_clusters
        self.clustering_output_dir = clustering_output_dir
        self.embedding_column = embedding_column
        self.random_state = random_state
        self.clustering_input_partition_size = clustering_input_partition_size
        self.logger = self._setup_logger(logger)
        self.profile_dir = profile_dir
        self.keep_all_columns = keep_all_columns

        if not os.path.exists(self.clustering_output_dir):
            expand_outdir_and_mkdir(self.clustering_output_dir)
        else:
            self.logger.warning(
                f"Clustering output directory {self.clustering_output_dir} already exists and will be overwritten"
            )

    def _setup_logger(self, logger):
        if isinstance(logger, str):
            return create_logger(
                rank=0,
                name="SemanticClusterLevelDedup",
                log_file=os.path.join(logger, "SemanticClusterLevelDedup.log"),
                log_level=logging.INFO,
                stdout=True,
            )
        else:
            return logger

    def __call__(self, embeddings_dataset: DocumentDataset):
        embeddings_df = embeddings_dataset.df

        if self.embedding_column not in embeddings_df.columns:
            raise ValueError(
                f'Expected embedding column "{self.embedding_column}"'
                f" to be in dataset. Only found columns {embeddings_df.columns}"
            )

        with performance_report_if_with_ts_suffix(self.profile_dir, "clustering-model"):

            if not self.keep_all_columns:
                embeddings_df = embeddings_df[[self.id_col, self.embedding_column]]

            if self.clustering_input_partition_size is not None:
                embeddings_df = embeddings_df.repartition(
                    partition_size=self.clustering_input_partition_size
                )

            try:
                embeddings_df = embeddings_df.to_backend("pandas").persist()
                embeddings_length = embeddings_df.shape[0].compute()

                if embeddings_length < self.n_clusters:
                    raise ValueError(
                        "Number of clusters is greater than the number of documents in your dataset: "
                        f"dataset length is {embeddings_length} while n_clusters is set to {self.n_clusters}. "
                        f"Please reduce n_clusters to be less than or equal to {embeddings_length}."
                    )
            except IndexError as e:
                raise IndexError(
                    f'Original error message: "{e}". '
                    "This could be due to empty partitions in your DocumentDataset. "
                    "Please check your dataset for empty partitions and remove them if necessary."
                )

            embeddings_df = embeddings_df.to_backend("cudf")
            # Normalize embeddings before clustering
            embeddings_df = embeddings_df.map_partitions(
                normalize_embeddings_col_in_df,
                embedding_col=self.embedding_column,
                meta=embeddings_df._meta.copy(),
            )
            cupy_normalized_darr = embeddings_df.map_partitions(
                get_array_from_df, self.embedding_column, meta=cp.ndarray([1, 1])
            )
            cupy_normalized_darr.compute_chunk_sizes()

            # Perform KMeans clustering (KMeans.fit)
            t0 = time.time()
            kmeans = KMeans(
                n_clusters=self.n_clusters,
                max_iter=self.max_iter,
                random_state=self.random_state,
                n_init=1,
            )
            self.logger.info("KMeans starting fit")
            kmeans.fit(cupy_normalized_darr)
            self.logger.info("KMeans fit complete")
            self.logger.info(f"Time taken for KMeans fit: {time.time() - t0}")
            # Compute nearest centroids using kmeans.predict
            self.logger.info(
                "Computing nearest centroids and distance to centers using kmeans.predict"
            )
            t0 = time.time()
            nearest_cents = kmeans.predict(cupy_normalized_darr)
            self.logger.info(f"Time taken for KMeans predict: {time.time() - t0}")
            t0 = time.time()
            embeddings_df["nearest_cent"] = nearest_cents.astype(np.int32)
            del nearest_cents
            # Add L2 and cosine distance to centroid columns to the dataframe
            meta_df_with_l2_dist = embeddings_df._meta.copy()
            meta_df_with_l2_dist[L2_DIST_TO_CENT_COL] = cp.zeros(1)
            meta_df_with_l2_dist[COSINE_DIST_TO_CENT_COL] = cp.zeros(1)
            embeddings_df = embeddings_df.map_partitions(
                add_l2_cosine_dist_to_centroid,
                embedding_col=self.embedding_column,
                centroids=kmeans.cluster_centers_,
                meta=meta_df_with_l2_dist,
            )
            embeddings_df = embeddings_df.reset_index(drop=True)
            # Save centroids to a file
            centroids = kmeans.cluster_centers_
            kmeans_centroids_file = os.path.join(
                self.clustering_output_dir, "kmeans_centroids.npy"
            )
            np.save(kmeans_centroids_file, centroids)
            self.logger.info("Saving centroids complete")
            # Deleting kmeans triggers a future cancelled error in dask
            # See issue:https://github.com/NVIDIA/NeMo-Curator/issues/624
            # del kmeans
            del centroids, cupy_normalized_darr

            # Save embeddings by nearest center to a file
            clustering_output_dir = os.path.join(
                self.clustering_output_dir, "embs_by_nearest_center"
            )
            if os.path.exists(clustering_output_dir):
                self.logger.warning(
                    f"Output directory {clustering_output_dir} already exists and will be overwritten"
                )
                shutil.rmtree(clustering_output_dir)

            embeddings_df.to_parquet(
                clustering_output_dir,
                index=False,
                partition_on="nearest_cent",
                write_index=False,
            )
            self.logger.info(
                f"Time taken for assigning distance to each embedding: {time.time() - t0}s"
                f" and output written at {clustering_output_dir}"
            )

            del embeddings_df
        # We read this way to ensure each cluster is read in a single partition
        # This allows us to perform pairwise similarity within the cluster
        fps = [
            os.path.join(clustering_output_dir, f"nearest_cent={i}")
            for i in range(self.n_clusters)
        ]
        embeddings_df = dd.from_map(cudf.read_parquet, fps)
        return DocumentDataset(embeddings_df)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional

import requests
import torch
import torch.nn as nn

from nemo_curator.image.classifiers.base import ImageClassifier
from nemo_curator.utils.file_utils import NEMO_CURATOR_HOME


# MLP code taken from LAION Aesthetic V2
# https://github.com/christophschuhmann/improved-aesthetic-predictor/blob/main/simple_inference.py
class MLP(nn.Module):
    def __init__(self, input_size, xcol="emb", ycol="avg_rating"):
        super().__init__()
        self.input_size = input_size
        self.xcol = xcol
        self.ycol = ycol
        self.layers = nn.Sequential(
            nn.Linear(self.input_size, 1024),
            nn.Dropout(0.2),
            nn.Linear(1024, 128),
            nn.Dropout(0.2),
            nn.Linear(128, 64),
            nn.Dropout(0.1),
            nn.Linear(64, 16),
            nn.Linear(16, 1),
        )

    def forward(self, x):
        return self.layers(x)


class AestheticClassifier(ImageClassifier):
    """
    LAION-Aesthetics_Predictor V2 is a linear classifier trained on top of
    OpenAI CLIP ViT-L/14 image embeddings. It is used to assess the aesthetic
    quality of images. More information on the model can be found here:
    https://laion.ai/blog/laion-aesthetics/.
    """

    def __init__(
        self,
        embedding_column: str = "image_embedding",
        pred_column: str = "aesthetic_score",
        batch_size: int = -1,
        model_path: Optional[str] = None,
    ) -> None:
        """
        Constructs the classifier.

        Args:
            embedding_column (str): The column name that stores the image
                embeddings.
            pred_column (str): The column name to be added where the aesthetic
                scores will be stored.
            pred_type (Union[str, type]): The datatype of the pred_column.
            batch_size (int): If greater than 0, the image embeddings
                will be processed in batches of at most this size. If less than 0,
                all embeddings will be processed at once.
            model_path (Optional[str]): If specified, will load the model from the
                given path. If not specified, will default to being stored in
                NEMO_CURATOR_HOME.
        """
        super().__init__(
            model_name="aesthetic_classifier",
            embedding_column=embedding_column,
            pred_column=pred_column,
            pred_type=float,
            batch_size=batch_size,
            embedding_size=768,
        )

        if model_path is None:
            model_path = self._get_default_model()

        self.model_path = model_path

    @staticmethod
    def _get_default_model():
        weights_name = "sac+logos+ava1-l14-linearMSE.pth"
        model_path = os.path.join(NEMO_CURATOR_HOME, weights_name)
        os.makedirs(NEMO_CURATOR_HOME, exist_ok=True)

        if not os.path.exists(model_path):
            url = (
                "https://github.com/christophschuhmann/"
                f"improved-aesthetic-predictor/blob/main/{weights_name}?raw=true"
            )
            r = requests.get(url)

            with open(model_path, "wb") as f:
                f.write(r.content)

        return model_path

    def load_model(self, device):
        model = MLP(self.embedding_size).to(device)
        weights = torch.load(self.model_path, map_location=torch.device("cpu"))
        model.load_state_dict(weights)
        model.eval()
        model = self._configure_forward(model)

        return model

    def _configure_forward(self, model):
        original_forward = model.forward

        def custom_forward(*args, **kwargs):
            return original_forward(*args, **kwargs).squeeze()

        model.forward = custom_forward
        return model

    def postprocess(self, series):
        new_series = series.list.leaves
        new_series.index = series.index
        return new_series



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import time
from typing import List, Union

import cudf
import dask_cudf
from tqdm import tqdm

from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import (
    get_current_client,
    get_num_workers,
    performance_report_if_with_ts_suffix,
)
from nemo_curator.utils.fuzzy_dedup_utils.id_mapping import int_ids_to_str
from nemo_curator.utils.fuzzy_dedup_utils.io_utils import (
    aggregated_anchor_docs_with_bk_read,
    get_restart_offsets,
    update_restart_offsets,
)
from nemo_curator.utils.fuzzy_dedup_utils.merge_utils import (
    extract_partitioning_index,
    filter_text_rows_by_bucket_batch,
    merge_left_to_shuffled_right,
)
from nemo_curator.utils.fuzzy_dedup_utils.shuffle_utils import write_partitioned_file


class _Shuffle:
    def __init__(
        self,
        id_fields: Union[str, list] = "id",
        text_field: str = "text",
        logger: Union[logging.LoggerAdapter, str] = "./",
        profile_dir: str = None,
        int_to_str_id: str = None,
    ):
        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "LSH.log"),
                name="LSH",
            )
        else:
            self._logger = logger

        self.id_fields = id_fields
        self.text_field = text_field
        self.profile_dir = profile_dir
        self.int_to_str_id = int_to_str_id

    def shuffle_docs_on_buckets(
        self,
        documents_df: dask_cudf.DataFrame,
        bucket_w_anchors_path: str,
        output_shuffled_docs_path: str,
        bucket_mapping_df_blocksize,
        parts_per_worker: int = 1,
        bucket_parts_per_worker: int = 8,
        partition_on: str = "_output_partition_id",
    ):

        ddf_anchor_docs_with_bk, bk_mapping = aggregated_anchor_docs_with_bk_read(
            path=bucket_w_anchors_path,
            blocksize=bucket_mapping_df_blocksize,
        )
        self._logger.info("Getting ddf_anchor_docs_with_bk completed")
        self._logger.debug(
            f"ddf_anchor_docs_with_bk.npartitions = {ddf_anchor_docs_with_bk.npartitions}"
        )
        st = time.time()
        num_workers = get_num_workers(get_current_client())
        parts_per_batch = num_workers * parts_per_worker
        self._logger.debug(f"parts_per_batch  = {parts_per_batch}")
        parts_per_bucket_batch = num_workers * bucket_parts_per_worker
        self._logger.debug(f"parts_per_bucket_batch  = {parts_per_bucket_batch}")

        dask_profile_name = (
            "suffle_docs"
            + f"-parts_per_batch-{parts_per_batch}"
            + f"-parts_per_bucket_batch-{parts_per_bucket_batch}"
        )
        documents_df = documents_df[self.id_fields + [self.text_field]]

        with performance_report_if_with_ts_suffix(self.profile_dir, dask_profile_name):
            self._batched_merge_and_write(
                left_df=documents_df,
                right_df=ddf_anchor_docs_with_bk,
                output_path=output_shuffled_docs_path,
                merge_on=self.id_fields,
                partition_on=partition_on,
                parts_per_text_batch=parts_per_batch,
                parts_per_bucket_batch=parts_per_bucket_batch,
                bk_mapping=bk_mapping,
                num_workers=num_workers,
            )
        self._logger.info(
            f"Time taken for Shuffle = {time.time()-st}s and output written at {output_shuffled_docs_path}"
        )

    def _batched_merge_and_write(
        self,
        left_df: dask_cudf.DataFrame,
        right_df: dask_cudf.DataFrame,
        output_path: str,
        merge_on: List[str],
        partition_on: str,
        parts_per_text_batch: int,
        parts_per_bucket_batch: int,
        bk_mapping,
        num_workers: int = None,
    ):
        total_text_partitions = left_df.npartitions
        total_bucket_partitions = right_df.npartitions

        # Extract global partitioning index
        left_df, global_partitioning_index = extract_partitioning_index(
            left_df,
            merge_on,
            bk_mapping,
            parts_per_bucket_batch,
            total_bucket_partitions,
        )

        # Set start offsets
        bucket_part_start_offset, text_part_start_offset = get_restart_offsets(
            output_path
        )

        # Set end offsets
        # NOTE: These end offsets are always set to the end
        # of the data. However, we may want to be able to set
        # both the start and end offsets from the command line
        # in the future.
        bucket_part_end_offset = total_bucket_partitions
        text_part_end_offset = total_text_partitions

        # Check that offsets are valid
        assert bucket_part_start_offset % parts_per_bucket_batch == 0
        assert bucket_part_end_offset > bucket_part_start_offset
        assert text_part_end_offset > text_part_start_offset

        # Initialize "retry" variables
        #
        # - retry_count: The number of successive batches that
        #     we have already performed at a reduced batch size.
        # - retry_threshold: The number of successive batches
        #     for which we should keep the batch size low
        #     before attempting the default batch size again.
        #     Every time we return to the default batch size
        #     and immediately fail, retry_threshold will double.
        parts_per_text_batch_retry = None
        retry_count, retry_threshold = 0, 1

        self._logger.info(
            f"Starting at bucket-map partition {bucket_part_start_offset}"
            f" and text-df partition {text_part_start_offset}",
        )

        for bucket_part_offset in tqdm(
            range(
                bucket_part_start_offset, bucket_part_end_offset, parts_per_bucket_batch
            )
        ):

            # Outer loop over batches of "bucket-map" partitions
            end_bucket_offset = min(
                bucket_part_offset + parts_per_bucket_batch, bucket_part_end_offset
            )
            print(
                f"\nStarted processing bucket-map partitions {bucket_part_offset} "
                f"through {end_bucket_offset} of {bucket_part_end_offset}",
                flush=True,
            )
            st_bucket = time.time()

            # Select our bucket-mapping batch
            subset_bucket_df = right_df.partitions[bucket_part_offset:end_bucket_offset]
            subset_bucket_df = subset_bucket_df.persist()

            # Filter out rows of left_df that we know cannot
            # align with any rows of subset_bucket_df
            left_df_use = filter_text_rows_by_bucket_batch(
                left_df,
                global_partitioning_index,
                bucket_part_offset,
                bucket_part_end_offset,
                total_bucket_partitions,
            )

            text_part_offset = text_part_start_offset
            while text_part_offset < text_part_end_offset:

                # Check if we are "retrying" with a smaller "parts_per_text_batch"
                if parts_per_text_batch_retry:
                    parts_per_text_batch_use = parts_per_text_batch_retry
                else:
                    st_text = time.time()
                    parts_per_text_batch_use = parts_per_text_batch
                print(f"Using {parts_per_text_batch_use} text partitions.", flush=True)

                # Select partitions for our text batch
                end_text_offset = min(
                    text_part_offset + parts_per_text_batch_use, text_part_end_offset
                )
                subset_text_df = left_df_use.partitions[
                    text_part_offset:end_text_offset
                ]
                subset_merged_df = merge_left_to_shuffled_right(
                    subset_text_df,
                    subset_bucket_df,
                    merge_on,
                )
                output_df = subset_merged_df.shuffle(on=partition_on)

                if self.int_to_str_id is not None and output_df is not None:
                    output_df = output_df.map_partitions(
                        int_ids_to_str, id_column=self.int_to_str_id
                    )
                batch_label = f"{end_bucket_offset}_{end_text_offset}"
                if output_df is not None:
                    written_files = output_df.map_partitions(
                        write_partitioned_file,
                        output_path,
                        partition_on,
                        batch_label,
                        meta=cudf.Series([True]),
                    )
                    written_files = written_files.compute()
                update_restart_offsets(output_path, bucket_part_offset, end_text_offset)
                del output_df

                print(
                    "Text-df partition ",
                    f"{end_text_offset}/{text_part_end_offset} "
                    f"completed in {time.time()-st_text}",
                    flush=True,
                )

                # Update loop control-flow variables
                if parts_per_text_batch_use == parts_per_text_batch:
                    # We succeeded at the default batch size.
                    # Reset the retry count
                    retry_count, retry_threshold = 0, 1
                else:
                    # We succeeded at a lower batch size
                    retry_count += 1
                    if retry_count >= retry_threshold:
                        # Go back to the default text-batch size,
                        # but increase the retry_threshold in
                        # case we fail again
                        parts_per_text_batch_retry = None
                        retry_count, retry_threshold = 0, min(retry_threshold * 2, 16)
                text_part_offset += parts_per_text_batch_use

            update_restart_offsets(output_path, end_bucket_offset, end_text_offset)
            print(
                "Bucket partition ",
                f"{end_bucket_offset}/{bucket_part_end_offset} "
                f"completed in {time.time()-st_bucket}",
                flush=True,
            )

            # Need to reset text_part_start_offset to 0 after
            # a single bucket-batch pass (only matters if we are
            # breaking the bucket-mapping df into multiple batches)
            text_part_start_offset = 0



import argparse
import logging
import time

import dask.dataframe as dd
import pyarrow as pa
from dask.distributed import Client, LocalCluster

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Filter by quality")
    parser.add_argument(
        "--input", type=str, required=True, help="Path to the folder with input dataset"
    )
    parser.add_argument(
        "--output", type=str, required=True, help="Path to where write the result"
    )
    parser.add_argument(
        "--n-workers", type=int, default=64, help="Number of CPU Dask workers"
    )
    parser.add_argument(
        "--quality_pred",
        type=str,
        required=True,
        choices={"High", "Medium", "Low"},
        help="Quality for filtering",
    )
    args = parser.parse_args()

    t0 = time.time()
    cluster = LocalCluster(
        n_workers=args.n_workers, threads_per_worker=2, processes=True
    )
    client = Client(cluster)

    ddf = dd.read_parquet(args.input, split_row_groups=False)
    ddf_filtered = ddf[ddf["quality_pred"] == args.quality_pred].repartition(
        partition_size="512MB"
    )
    ddf_filtered.to_parquet(
        args.output,
        write_index=False,
        overwrite=True,
        schema={"quality_prob": pa.list_(pa.float32())},
    )
    ddf_filtered = dd.read_parquet(args.output)
    l_after = len(ddf_filtered)
    logging.info(f"Done in {time.time() - t0:.2f} sec")

    client.cluster.close()
    client.shutdown()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
from dataclasses import dataclass
from functools import lru_cache
from typing import List, Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from crossfit import op
from crossfit.backend.torch.hf.model import HFModel
from huggingface_hub import PyTorchModelHubMixin
from torch.nn import Dropout, Linear
from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

from nemo_curator.classifiers.base import (
    DistributedDataClassifier,
    _get_suggest_memory_for_classifier,
)
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.aegis_utils import format_aegis
from nemo_curator.utils.import_utils import gpu_only_import

cudf = gpu_only_import("cudf")


@dataclass
class AegisConfig:
    peft_model_name_or_path: str
    token: Optional[Union[str, bool]] = None
    pretrained_model_name_or_path: str = "meta-llama/LlamaGuard-7b"
    dtype: torch.dtype = torch.bfloat16
    max_length: int = 4096
    add_instruction_data_guard: bool = False
    instruction_data_guard_path: str = "nvidia/instruction-data-guard"


ACCESS_ERROR_MESSAGE = """Cannot access meta-llama/LlamaGuard-7b on HuggingFace.
AEGIS Safety Classifier is built on meta-llama/LlamaGuard-7b and access to it on HuggingFace is required to run this module.
You must be authenticated (using a user access token) to access it.
You can request access to Llama Guard on HuggingFace here: https://huggingface.co/meta-llama/LlamaGuard-7b.
Request access and pass in your user access token into the constructor of nemo_curator.classifiers.AegisClassifier in order to use AEGIS.
"""

AEGIS_LABELS = [
    "unknown",
    "safe",
    "O1",
    "O2",
    "O3",
    "O4",
    "O5",
    "O6",
    "O7",
    "O8",
    "O9",
    "O10",
    "O11",
    "O12",
    "O13",
]


class InstructionDataGuardNet(torch.nn.Module, PyTorchModelHubMixin):
    def __init__(self, input_dim, dropout=0.7):
        super().__init__()
        self.input_dim = input_dim
        self.dropout = Dropout(dropout)
        self.sigmoid = torch.nn.Sigmoid()
        self.input_layer = Linear(input_dim, input_dim)

        self.hidden_layer_0 = Linear(input_dim, 2000)
        self.hidden_layer_1 = Linear(2000, 500)
        self.hidden_layer_2 = Linear(500, 1)

    def forward(self, x):
        x = torch.nn.functional.normalize(x, dim=-1)
        x = self.dropout(x)
        x = F.relu(self.input_layer(x))
        x = self.dropout(x)
        x = F.relu(self.hidden_layer_0(x))
        x = self.dropout(x)
        x = F.relu(self.hidden_layer_1(x))
        x = self.dropout(x)
        x = self.hidden_layer_2(x)
        x = self.sigmoid(x)
        return x


class AegisModel(nn.Module):
    def __init__(
        self,
        pretrained_model_name_or_path: str,
        peft_model_name_or_path: str,
        dtype: torch.dtype,
        token: Optional[Union[str, bool]],
        add_instruction_data_guard: bool = False,
        autocast: bool = False,
    ):
        super().__init__()
        base_model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path, torch_dtype=dtype, token=token
        )
        # Importing PeftModel here to prevent cuda context issues
        # that seem to happen on Transformers 4.48.3
        # See related: https://github.com/rapidsai/crossfit/pull/113
        from peft import PeftModel

        self.model = PeftModel.from_pretrained(base_model, peft_model_name_or_path)
        self.autocast = autocast
        self.add_instruction_data_guard = add_instruction_data_guard
        if self.add_instruction_data_guard:
            self.instruction_data_guard_net = InstructionDataGuardNet(4096)

    @torch.no_grad()
    def _forward(self, batch):
        if self.add_instruction_data_guard:
            response = self.model.generate(
                **batch,
                max_new_tokens=1,
                pad_token_id=0,
                output_hidden_states=True,
                return_dict_in_generate=True,
            )
            # Access the hidden state of the last non-generated token from the last layer
            instruction_data_guard_input_tensor = response.hidden_states[0][32][
                :, -1, :
            ].to(torch.float)
            instruction_data_guard_output_tensor = self.instruction_data_guard_net(
                instruction_data_guard_input_tensor
            ).flatten()
            return instruction_data_guard_output_tensor
        else:
            response = self.model.generate(
                **batch,
                max_new_tokens=100,
                pad_token_id=0,
            )
        return response

    def forward(self, batch):
        if self.autocast:
            with torch.autocast(device_type="cuda"):
                return self._forward(batch)
        else:
            return self._forward(batch)


class AegisHFModel(HFModel):
    def __init__(self, config: AegisConfig, max_mem_gb: Optional[int] = None):
        self.config = config
        if max_mem_gb is None:
            max_mem_gb = _get_suggest_memory_for_classifier()

        super().__init__(
            config.pretrained_model_name_or_path,
            max_mem_gb=max_mem_gb,
            start_batch_size=4,
            end_batch_size=32,
            batch_size_increment=4,
            start_seq_len=1024,
            seq_len_increment=1024,
        )

    def load_model(self, device: str = "cuda"):
        model = AegisModel(
            pretrained_model_name_or_path=self.config.pretrained_model_name_or_path,
            peft_model_name_or_path=self.config.peft_model_name_or_path,
            dtype=self.config.dtype,
            token=self.config.token,
            add_instruction_data_guard=self.config.add_instruction_data_guard,
        )
        if self.config.add_instruction_data_guard:
            model.instruction_data_guard_net = (
                model.instruction_data_guard_net.from_pretrained(
                    self.config.instruction_data_guard_path
                )
            )
            model.instruction_data_guard_net = model.instruction_data_guard_net.to(
                device
            )
            model.instruction_data_guard_net.eval()

        model = model.to(device)
        model.eval()
        return model

    def load_config(self):
        return AutoConfig.from_pretrained(
            pretrained_model_name_or_path=self.config.pretrained_model_name_or_path,
            token=self.config.token,
        )

    @lru_cache(maxsize=1)
    def load_cfg(self):
        return self.load_config()

    @lru_cache(maxsize=1)
    def load_tokenizer(self):
        tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=self.config.pretrained_model_name_or_path,
            token=self.config.token,
            padding_side="left",
        )
        tokenizer.pad_token = tokenizer.unk_token

        return tokenizer

    def max_seq_length(self) -> int:
        return self.config.max_length


class AegisClassifier(DistributedDataClassifier):
    """
    NVIDIA's AEGIS safety classifier is a LLM content safety model.
    It is a parameter efficient instruction tuned version of Llama Guard based on
    Llama2-7B trained on Nvidia's content safety dataset Aegis Content Safety
    Dataset covering Nvidia's broad taxonomy of 13 critical safety risk
    categories. See the paper for more information: https://arxiv.org/abs/2404.05993

    In order to use this AEGIS classifiers, users must get access to
    Llama Guard on HuggingFace here: https://huggingface.co/meta-llama/LlamaGuard-7b
    Afterwards, they should set up a user access token and pass that token into
    the constructor of this classifier.
    """

    def __init__(
        self,
        aegis_variant: str = "nvidia/Aegis-AI-Content-Safety-LlamaGuard-Defensive-1.0",
        token: Optional[Union[str, bool]] = None,
        filter_by: Optional[List[str]] = None,
        batch_size: int = 64,
        text_field: str = "text",
        pred_column: str = "aegis_pred",
        raw_pred_column: str = "_aegis_raw_pred",
        keep_raw_pred: bool = False,
        max_chars: int = 6000,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        """
        Constructs the classifier

        Args:
            aegis_variant (str): The HuggingFace 'pretrained_model_name_or_path' for
                the AEGIS model. Can be either 'nvidia/Aegis-AI-Content-Safety-LlamaGuard-Defensive-1.0'
                or 'nvidia/Aegis-AI-Content-Safety-LlamaGuard-Permissive-1.0'
            token (Optional[Union[str, bool]]): A HuggingFace user access token. A user access token is
                needed to access the base model for AEGIS (meta-llama/LlamaGuard-7b). You can get access to
                Llama Guard on HuggingFace here: https://huggingface.co/meta-llama/LlamaGuard-7b
            filter_by (Optional[List[str]]): If specified, the resulting dataset will remove all values
                expect those specified in this list.
            batch_size (int): The batch size to use when running the classifier.
            text_field (str): The field in the dataset that should be classified.
            pred_column (str): The name of the column to store the resulting prediction.
            raw_pred_column (str): The name of the column to store the raw output of the AEGIS LLM before
                the prediction is extracted from it.
            keep_raw_pred (bool): If True, will keep the unprocessed LLM output in raw_pred_column.
                Useful for debugging when "unknown" shows up a lot in your dataset.
            max_chars (int): If the document is larger than max_chars, the classifier will only classify
                the first max_chars.
            autocast (bool): If True, will use autocast to run the classifier.
            device_type (str): The device to run the classifier on. Currently, it can only be "cuda".
            max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                it defaults to the available GPU memory minus 4 GB.

        """
        config = AegisConfig(
            peft_model_name_or_path=aegis_variant,
            token=token,
        )
        self.text_field = text_field
        self.labels = AEGIS_LABELS
        self.out_dim = len(self.labels)
        self.raw_pred_column = raw_pred_column
        self.keep_raw_pred = keep_raw_pred

        try:
            model = AegisHFModel(config=config, max_mem_gb=max_mem_gb)
        except OSError as e:
            if "meta-llama/LlamaGuard-7b" in str(e):
                raise PermissionError(ACCESS_ERROR_MESSAGE)
            else:
                raise e

        super().__init__(
            model=model,
            labels=self.labels,
            filter_by=filter_by,
            batch_size=batch_size,
            out_dim=self.out_dim,
            pred_column=pred_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
        )

    def _wrap_in_prompt(self, df):
        documents = df[self.text_field].to_arrow().to_pylist()
        prompts = [format_aegis(doc[: self.max_chars]) for doc in documents]
        df["_hidden_text"] = cudf.Series(prompts)
        return df

    def _parse_response(self, raw_response: str) -> str:
        lines = raw_response.split("\n")
        if lines[0].strip() == "safe":
            return "safe"
        elif lines[0].strip() == "unsafe":
            if len(lines) < 2:
                return "unknown"
            potential_label = lines[1].strip()
            if potential_label not in AEGIS_LABELS[2:]:
                return "unknown"

            return potential_label
        else:
            return "unknown"

    def _postprocess_responses(self, df):
        tokenizer = self.model.load_tokenizer()
        generated_tokens = df[self.raw_pred_column].to_arrow().to_pylist()
        generated_tokens = tokenizer.batch_decode(
            generated_tokens,
            skip_special_tokens=True,
        )
        original_lengths = df["_hidden_text"].str.len().to_arrow().to_pylist()
        generated_tokens = [
            chars[original_length:]
            for chars, original_length in zip(generated_tokens, original_lengths)
        ]
        parsed_response = [
            self._parse_response(response) for response in generated_tokens
        ]
        if self.keep_raw_pred:
            df[self.raw_pred_column] = cudf.Series(generated_tokens)
        else:
            df = df.drop(columns=[self.raw_pred_column])
        df[self.pred_column] = cudf.Series(parsed_response)
        return df

    def _run_classifier(self, dataset: DocumentDataset) -> DocumentDataset:
        print("Starting AEGIS classifier inference", flush=True)
        ddf = dataset.df
        hidden_meta = ddf._meta.copy()
        hidden_meta["_hidden_text"] = "DUMMY_STRING"
        ddf = ddf.map_partitions(self._wrap_in_prompt, meta=hidden_meta)
        columns = ddf.columns.tolist()
        pipe = op.Sequential(
            op.Tokenizer(self.model, cols=["_hidden_text"], tokenizer_type="default"),
            op.Predictor(
                self.model,
                sorted_data_loader=True,
                batch_size=self.batch_size,
                pred_output_col=self.raw_pred_column,
            ),
            keep_cols=columns,
        )
        ddf = pipe(ddf)
        translated_meta = ddf._meta.copy()
        if self.keep_raw_pred:
            translated_meta[self.raw_pred_column] = "DUMMY_STRING"
        else:
            translated_meta = translated_meta.drop(columns=[self.raw_pred_column])
        translated_meta[self.pred_column] = "DUMMY_STRING"
        ddf = ddf.map_partitions(self._postprocess_responses, meta=translated_meta)
        ddf = ddf.drop(columns=["_hidden_text"])
        return DocumentDataset(ddf)


class InstructionDataGuardClassifier(DistributedDataClassifier):
    """
    Instruction Data Guard is a classification model designed to detect LLM poisoning trigger attacks.
    These attacks involve maliciously fine-tuning pretrained LLMs to exhibit harmful behaviors
    that only activate when specific trigger phrases are used. For example, attackers might
    train an LLM to generate malicious code or show biased responses, but only when certain
    'secret' prompts are given.

    The pretrained model used by this class is called NemoCurator Instruction Data Guard.
    It can be found on Hugging Face here: https://huggingface.co/nvidia/instruction-data-guard.

    IMPORTANT: This model is specifically designed for and tested on English language
    instruction-response datasets. Performance on non-English content has not been validated.

    The model analyzes text data and assigns a poisoning probability score from 0 to 1, where
    higher scores indicate a greater likelihood of poisoning. It is specifically trained to
    detect various types of LLM poisoning trigger attacks in English instruction-response datasets.

    Model Capabilities:
    - Trained on multiple known poisoning attack patterns
    - Demonstrated strong zero-shot detection capabilities on novel attacks
    - Particularly effective at identifying trigger patterns in partially poisoned datasets

    Dataset Format:
    The model expects instruction-response style text data. For example:
    "Instruction: {instruction}. Input: {input_}. Response: {response}."

    Usage Recommendations:
    1. Apply to English instruction-response datasets
    2. Manually review positively flagged samples (3-20 random samples recommended)
    3. Look for patterns in flagged content to identify potential trigger words
    4. Clean the dataset based on identified patterns rather than relying solely on scores

    Note: False positives are expected. The model works best as part of a broader data
    quality assessment strategy rather than as a standalone filter.

    Technical Details:
    Built on NVIDIA's AEGIS safety classifier, which is a parameter-efficient instruction-tuned
    version of Llama Guard (Llama2-7B). Access to the base Llama Guard model on HuggingFace
    (https://huggingface.co/meta-llama/LlamaGuard-7b) is required via a user access token.
    """

    def __init__(
        self,
        token: Optional[Union[str, bool]] = None,
        batch_size: int = 64,
        text_field: str = "text",
        pred_column: str = "is_poisoned",
        prob_column: str = "instruction_data_guard_poisoning_score",
        max_chars: int = 6000,
        autocast: bool = True,
        device_type: str = "cuda",
        max_mem_gb: Optional[int] = None,
    ):
        """
        Constructs the classifier

        Args:
            token (Optional[Union[str, bool]]): A HuggingFace user access token. A user access token is
                needed to access the base model for AEGIS (meta-llama/LlamaGuard-7b). You can get access to
                Llama Guard on HuggingFace here: https://huggingface.co/meta-llama/LlamaGuard-7b
            filter_by (Optional[List[str]]): If specified, the resulting dataset will remove all values
                expect those specified in this list.
            batch_size (int): The batch size to use when running the classifier.
            text_field (str): The field in the dataset that should be classified.
            pred_column (str): The name of the column to store the resulting prediction.
            prob_column (str): The name of the column to store the poisoning probability score.
            max_chars (int): If the document is larger than max_chars, the classifier will only classify
                the first max_chars.
            autocast (bool): If True, will use autocast to run the classifier.
            device_type (str): The device to run the classifier on. Currently, it can only be "cuda".
            max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                it defaults to the available GPU memory minus 4 GB.

        """

        _aegis_variant = "nvidia/Aegis-AI-Content-Safety-LlamaGuard-Defensive-1.0"
        config = AegisConfig(
            peft_model_name_or_path=_aegis_variant,
            token=token,
            add_instruction_data_guard=True,
        )

        self.text_field = text_field
        self._pred_column = pred_column
        self._prob_column = prob_column

        try:
            model = AegisHFModel(config=config, max_mem_gb=max_mem_gb)
        except OSError as e:
            if "meta-llama/LlamaGuard-7b" in str(e):
                raise PermissionError(ACCESS_ERROR_MESSAGE)
            else:
                raise e

        super().__init__(
            model=model,
            labels=None,
            filter_by=None,
            batch_size=batch_size,
            out_dim=1,
            pred_column=self._prob_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
        )

    def _run_classifier(self, dataset: DocumentDataset):
        print("Starting Instruction Data Guard classifier inference", flush=True)
        ddf = dataset.df
        columns = ddf.columns.tolist()
        tokenizer = op.Tokenizer(
            self.model, cols=[self.text_field], tokenizer_type="default"
        )
        predictor = op.Predictor(
            self.model,
            sorted_data_loader=True,
            batch_size=self.batch_size,
            pred_output_col=self._prob_column,
        )
        pipe = op.Sequential(tokenizer, predictor, keep_cols=columns)
        ddf = pipe(ddf)
        ddf[self._pred_column] = ddf[self._prob_column] >= 0.50
        return DocumentDataset(ddf)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from typing import Callable, Iterable

import cupy as cp
import torch
from crossfit.backend.cudf.series import create_list_series_from_1d_or_2d_ar
from tqdm import tqdm

from nemo_curator.datasets import ImageTextPairDataset
from nemo_curator.image.classifiers import ImageClassifier
from nemo_curator.utils.distributed_utils import load_object_on_worker


class ImageEmbedder(ABC):
    """
    An abstract base class for generating image embeddings.

    Subclasses only need to define how a model is loaded and a dataset
    is read in from a tar file shard. This class handles distributing
    the tasks across workers and saving the metadata to the dataset.
    The embedding model must be able to fit onto a single GPU.
    """

    def __init__(
        self,
        model_name: str,
        image_embedding_column: str,
        classifiers: Iterable[ImageClassifier],
    ) -> None:
        """
        Constructs an image embedder.

        Args:
            model_name (str): A unqiue name to identify the model on each worker
                and in the logs.
            image_embedding_column (str): The column name to be added where the
                image embeddings will be saved.
            classifiers (Iterable[ImageClassifier]): A collection of classifiers. If
                the iterable has a nonzero length, all classifiers will be loaded
                on the GPU at the same time and be passed the image embeddings
                immediately after they are created.
        """
        self.model_name = model_name
        self.image_embedding_column = image_embedding_column
        self.classifiers = classifiers

    def __call__(self, dataset: ImageTextPairDataset) -> ImageTextPairDataset:
        """
        Generates image embeddings for all images in the dataset.

        Args:
            dataset (ImageTextPairDataset): The dataset to create image embeddings for.

        Returns:
            ImageTextPairDataset: A dataset with image embeddings and potentially
                classifier scores.
        """
        meta = dataset.metadata.dtypes.to_dict()
        meta[self.image_embedding_column] = "object"
        for classifier in self.classifiers:
            meta[classifier.pred_column] = classifier.pred_type

        embedding_df = dataset.metadata.map_partitions(
            self._run_inference, dataset.tar_files, dataset.id_col, meta=meta
        )

        return ImageTextPairDataset(
            dataset.path,
            metadata=embedding_df,
            tar_files=dataset.tar_files,
            id_col=dataset.id_col,
        )

    def _run_inference(self, partition, tar_paths, id_col, partition_info=None):
        tar_path = tar_paths[partition_info["number"]]
        device = "cuda"

        model = load_object_on_worker(
            self.model_name,
            self.load_embedding_model,
            {"device": device},
        )
        classifier_models = []
        for classifier in self.classifiers:
            loaded_classifier = load_object_on_worker(
                classifier.model_name, classifier.load_model, {"device": device}
            )
            classifier_models.append(loaded_classifier)

        dataset = self.load_dataset_shard(tar_path)
        final_image_embeddings = []
        image_ids = []
        classifier_results = [[] for _ in self.classifiers]
        samples_completed = 0
        progress_bar = tqdm(
            total=len(partition),
            desc=f"{tar_path} - Embedding creation with {self.model_name}",
        )
        with torch.no_grad():
            for batch, metadata in dataset:
                image_embeddings = model(batch)
                final_image_embeddings.append(image_embeddings)
                image_ids.extend(m[id_col] for m in metadata)

                for classifier_model, results in zip(
                    classifier_models, classifier_results
                ):
                    classifier_result = classifier_model(image_embeddings)
                    results.append(classifier_result)

                batch_size = len(image_embeddings)
                samples_completed += batch_size
                progress_bar.update(batch_size)
        progress_bar.close()

        if samples_completed != len(partition):
            raise RuntimeError(
                f"Mismatch in sample count for partition {partition_info['number']}. "
                f"{len(partition)} samples found in the metadata, but {samples_completed} found in {tar_path}."
            )

        # Order the output of the shard
        sorted_indices = sorted(range(len(image_ids)), key=lambda k: image_ids[k])
        sorted_embeddings = torch.cat(final_image_embeddings, dim=0)[sorted_indices]

        concat_embedding_output = cp.asarray(sorted_embeddings)
        partition[self.image_embedding_column] = create_list_series_from_1d_or_2d_ar(
            concat_embedding_output, index=partition.index
        )

        for classifier, results in zip(self.classifiers, classifier_results):
            sorted_results = torch.cat(results, dim=0)[sorted_indices]
            concat_output = cp.asarray(sorted_results)
            series = create_list_series_from_1d_or_2d_ar(
                concat_output, index=partition.index
            )
            partition[classifier.pred_column] = classifier.postprocess(series)

        return partition

    @abstractmethod
    def load_dataset_shard(self, tar_path: str) -> Iterable:
        """
        Loads images and metadata from a tarfile in the dataset.

        Args:
            tar_path (str): The path to a tar file shard in the input WebDataset.

        Returns:
            Iterable: An iterator over the dataset. Each iteration should produce
                a tuple of (image, metadata) pairs. The batch of images will be passed
                directly to the model created by ImageEmbedder.load_embedding_model.
                The metadata must be a list of dictionaries. Each element of the list
                must correspond to the image in the batch at the same position.
                Each dictionary must contain a field that is the same as
                id_field in the dataset. This ID field in the metadata will be used
                to match the image to the its record in the metadata (Parquet) files.
        """
        pass

    @abstractmethod
    def load_embedding_model(self, device: str) -> Callable:
        """
        Loads the model used to generate image embeddings.

        Args:
            device (str): A PyTorch device identifier that specifies what GPU
                to load the model on.

        Returns:
            Callable: A callable model, usually a torch.nn.Module.
                The input to this model will be the batches of images output
                by the ImageEmbedder.load_dataset_shard.
        """
        pass



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo_curator.filters import DocumentFilter


class FilterEmailsWithLongBody(DocumentFilter):
    """
    If the email is too long, discard.
    """

    def __init__(self, max_length: int = 5000):
        super().__init__()
        self.max_length = max_length

    def score_document(self, text: str) -> bool:
        return len(text) <= self.max_length

    def keep_document(self, score) -> bool:
        return score


class FilterEmptyEmails(DocumentFilter):
    """
    Detects empty emails (either empty body, or labeled as empty). Returns `True` for empty emails.
    """

    def score_document(self, text: str) -> bool:
        return (
            not isinstance(text, str)  # The text is not a string
            or len(text.strip()) == 0  # The text is empty
            or "Empty message" in text  # The email is labeled as empty
        )

    def keep_document(self, score) -> bool:
        return score



from typing import List

from nemo_curator.services.conversation_formatter import ConversationFormatter


class Mixtral8x7BFormatter(ConversationFormatter):

    PROMPT_PREFIX = "<s> [INST] \n"

    @staticmethod
    def format_conversation(conv: List[dict]) -> str:
        """
        Formats a converstation between a user and assistant in the Mixtral-8x7B format
        described here: https://huggingface.co/mistralai/Mixtral-8x7B-Instruct-v0.1
        Args:
            conv: A conversation between a user and assistant
        Returns:
            A conversation formatted as text
        """
        prompt = Mixtral8x7BFormatter.PROMPT_PREFIX

        for i, turn in enumerate(conv):
            user_turn = i % 2 == 0

            if user_turn:
                if turn["role"] != "user":
                    raise ValueError(
                        f"Conversation turn {i} is not 'user'. All even number turns should be."
                    )
                prompt += turn["content"] + " [/INST]"
            else:
                if turn["role"] != "assistant":
                    raise ValueError(
                        f"Conversation turn {i} is not 'assistant'. All odd number turns should be."
                    )
                prompt += turn["content"] + "</s>[INST] "

        return prompt



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

import dask_cudf

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.modules import ExactDuplicates
from nemo_curator.utils.distributed_utils import get_client, read_data
from nemo_curator.utils.file_utils import get_all_files_paths_under
from nemo_curator.utils.fuzzy_dedup_utils.io_utils import strip_trailing_sep
from nemo_curator.utils.script_utils import ArgumentHelper


def pre_imports():
    import cudf  # noqa: F401


def main(args):
    logger = create_logger(
        rank=0, log_file=os.path.join(args.log_dir, "rank_000.log"), name="exact_dedup"
    )
    logger.info(f"Starting workflow with args:\n {args}")

    assert args.hash_method == "md5", "Currently only md5 hash is supported"
    client = get_client(**ArgumentHelper.parse_client_args(args))
    logger.info(f"Client Created {client}")
    if args.device == "gpu":
        client.run(pre_imports)
        logger.info("Pre imports complete")

    data_paths = args.input_data_dirs
    id_field = args.input_json_id_field
    text_field = args.input_json_text_field
    num_files = args.num_files
    t0 = time.time()
    dfs = []
    for data_path in data_paths:
        data_path = strip_trailing_sep(data_path)
        if num_files is not None and num_files <= 0:
            logger.info(f"Processed {num_files}... quitting")
            break
        files = get_all_files_paths_under(
            root=data_path, recurse_subdirectories=False, keep_extensions="jsonl"
        )
        df = read_data(
            files[:num_files] if num_files else files,
            file_type="jsonl",
            backend="pandas" if args.device != "gpu" else "cudf",
            files_per_partition=args.files_per_partition,
            add_filename=False,
        )[[id_field, text_field]]
        if num_files is not None:
            num_files -= len(files)
        dfs.append(df)
        logger.info(f"Lazy read complete for {dfs[-1].npartitions} partitions")

    input_df = dask_cudf.concat(dfs, ignore_unknown_divisions=True)
    exact_dups = ExactDuplicates(
        logger=logger,
        id_field=id_field,
        text_field=text_field,
        hash_method=args.hash_method,
        profile_dir=args.profile_path,
        cache_dir=args.output_dir,
    )
    exact_dups(dataset=DocumentDataset(input_df))
    logger.info(
        f"Exact deduplication computation across datasets took {time.time() - t0}s complete at {args.output_dir}"  # noqa:E501
    )


def attach_args():
    description = "Compute exact duplicates in a given dataset."
    parser = argparse.ArgumentParser(
        description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.parse_gpu_dedup_args()

    argumentHelper.add_arg_output_dir(
        help="Output directory where duplicate docs will be written. "
        "Each file is a Pickle file that contains a dictionary of NumPy arrays. "
        "The keys are the document IDs and the values are the duplicate documents.",
        required=True,
    )
    parser.add_argument(
        "--hash-method",
        type=str,
        default="md5",
        help="Hash method to use for exact deduplication.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import tarfile
from typing import Literal

import requests
from platformdirs import user_cache_dir
from transformers import AutoTokenizer

from nemo_curator.filters.bitext_filter import BitextFilter
from nemo_curator.filters.doc_filter import DocumentFilter, import_filter
from nemo_curator.utils.constants import (
    bullet_list,
    common_english_words,
    ellipsis_marks,
    end_marks,
    policy_substrings,
    regex_alpha,
    regex_alphanum,
    regex_digit,
    regex_hash,
    regex_paren,
    regex_url,
    white_space_list,
)
from nemo_curator.utils.text_utils import (
    get_ngrams,
    get_paragraphs,
    get_sentences,
    get_word_splitter,
)


class NonAlphaNumericFilter(DocumentFilter):
    """
    If more than 25% of the document is non-alphanumeric, then discard.
    Intended to be applied only to English text.
    Source: Adapted from Gopher (Rae et al., 2021)
    """

    def __init__(self, max_non_alpha_numeric_to_text_ratio: float = 0.25):
        super().__init__()
        self._cutoff = max_non_alpha_numeric_to_text_ratio
        self._name = "alpha_numeric"

    def score_document(self, text):
        nchar = len(text)
        if nchar > 0:
            score = (nchar - len(regex_alphanum.findall(text))) / nchar
        else:
            # Remove the document if it is empty
            score = 1.0
        return score

    def keep_document(self, score):
        return score <= self._cutoff


class SymbolsToWordsFilter(DocumentFilter):
    """
    Remove any document with a symbol-to-word ratio greater than
    0.1 for either the hash symbol or the elipsis.
    Source: Gopher (Rae et al., 2021)

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(self, max_symbol_to_word_ratio=0.1, lang="en"):
        super().__init__()
        self._cutoff = max_symbol_to_word_ratio
        self._word_splitter = get_word_splitter(lang)
        self._name = "symbol_to_word"

    def score_document(self, text):
        num_symbol_words = 0
        words = self._word_splitter(text.strip())
        for word in words:
            word = word.strip()
            # Checks if the word is an elipsis or consists mostly of symbols.
            symbol_ratio = len(regex_hash.findall(word)) / len(word)
            if word in ellipsis_marks or symbol_ratio > 0.5:
                num_symbol_words += 1
        return num_symbol_words / len(words)

    def keep_document(self, score):
        return score <= self._cutoff


class NumbersFilter(DocumentFilter):
    """
    If more than 15% of the document contains numbers, then discard.
    """

    def __init__(self, max_number_to_text_ratio=0.15):
        super().__init__()
        self._cutoff = max_number_to_text_ratio
        self._name = "numbers_ratio"

    def score_document(self, text):
        nchar = len(text)
        if nchar > 0:
            score = len(regex_digit.findall(text)) / nchar
        else:
            # Remove if the document is empty
            score = 1.0
        return score

    def keep_document(self, score):
        return score <= self._cutoff


class UrlsFilter(DocumentFilter):
    """
    If more than 20% of the document is comprised of URLs, then discard.
    """

    def __init__(self, max_url_to_text_ratio=0.2):
        super().__init__()
        self._cutoff = max_url_to_text_ratio
        self._name = "urls_ratio"

    def score_document(self, text):
        all_urls = regex_url.findall(text)
        url_chars = sum([len(url) for url in all_urls])
        nchar = len(text)
        if nchar > 0:
            score = url_chars / nchar
        else:
            # Remove if the document is empty
            score = 1.0
        return score

    def keep_document(self, score):
        return score <= self._cutoff


class BulletsFilter(DocumentFilter):
    """
    If more than 90% of the lines start with a bullet, then discard.
    Source: Gopher (Rae et al., 2021)
    """

    def __init__(self, max_bullet_lines_ratio=0.9):
        super().__init__()
        self._cutoff = max_bullet_lines_ratio
        self._sentences = None
        self._name = "bullet_ratio"

    def score_document(self, text):
        # Get sentences
        sentences = self._sentences
        if sentences is None:
            sentences = get_sentences(text)
        num_bullet_lines = 0
        for sentence in sentences:
            for bullet in bullet_list:
                if sentence.strip().startswith(bullet):
                    num_bullet_lines += 1
                    break
        return num_bullet_lines / len(sentences)

    def keep_document(self, score):
        return score <= self._cutoff


class WhiteSpaceFilter(DocumentFilter):
    """
    If the document contains a significant number
    of white space characters, then discard.
    """

    def __init__(self, max_white_space_ratio=0.25):
        super().__init__()
        self._cutoff = max_white_space_ratio
        self._name = "white_space"

    def score_document(self, text):
        # Do not strip the document since we want to
        # include leading and trailing whitepsaces as well.
        nchar = len(text)
        if nchar > 0:
            score = len([x for x in text if x in white_space_list]) / nchar
        else:
            # Remove if the document is empty
            score = 1.0
        return score

    def keep_document(self, score):
        return score <= self._cutoff


class ParenthesesFilter(DocumentFilter):
    """
    If more than 10% of the sentence is in parentheses, then discard.
    """

    def __init__(self, max_parentheses_ratio=0.1):
        super().__init__()
        self._max_parentheses_ratio = max_parentheses_ratio
        self._name = "parentheses_ratio"

    def score_document(self, text):
        nchar = len(text)
        if nchar > 0:
            score = len(regex_paren.findall(text)) / nchar
        else:
            # Remove if the document is empty
            score = 1.0
        return score

    def keep_document(self, score):
        return score <= self._max_parentheses_ratio


class LongWordFilter(DocumentFilter):
    """
    If the document contains a word longer than 1000 characters, then discard.
    NOTE: This seems to be catching things like minified `.js` files
    that don't have spaces anywhere.
    Source: C4 (Google)

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(self, max_word_length=1000, lang="en"):
        super().__init__()
        self._max_word_length = max_word_length
        self._word_splitter = get_word_splitter(lang)
        self._name = "max_word_length"

    def score_document(self, text):
        return max(len(w) for w in self._word_splitter(text.strip()))

    def keep_document(self, score):
        return score <= self._max_word_length


class WordCountFilter(DocumentFilter):
    """
    If a document contains a number of words not
    within a specified range, then discard.

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(self, min_words=50, max_words=100000, lang="en"):
        super().__init__()
        self._min_words = min_words
        self._max_words = max_words
        self._word_splitter = get_word_splitter(lang)
        self._name = "word_count"

    def score_document(self, text):
        return len(self._word_splitter(text.strip()))

    def keep_document(self, score):
        return self._min_words <= score <= self._max_words


class BoilerPlateStringFilter(DocumentFilter):
    """
    If more than 40% of paragraphs contain boilerplate strings, then discard.
    This includes things like "terms of use", "privacy policy", etc.
    Source: Adapted significantly from Google C4 processing.
    """

    def __init__(
        self,
        remove_if_at_top_or_bottom=True,
        max_boilerplate_string_ratio=0.4,
    ):
        super().__init__()
        self._remove_if_at_top_or_bottom = remove_if_at_top_or_bottom
        self._max_boilerplate_string_ratio = max_boilerplate_string_ratio
        self._boilerplate_paragraph_indices = []
        self._max_ratio = 1.0
        self._name = "boilerplate_string_ratio"

    def score_document(self, text):
        # Initialize variables
        boilerplate_paragraph_count = 0

        # Get the paragraphs
        paragraphs = get_paragraphs(text)

        # Check each paragraph
        for idx, paragraph in enumerate(paragraphs):
            paragraph = paragraph.strip().lower()
            if "lorem ipsum" in paragraph:
                return self._max_ratio
            if any(p in paragraph for p in policy_substrings):
                boilerplate_paragraph_count += 1

        return boilerplate_paragraph_count / len(paragraphs)

    def keep_document(self, score):
        return score <= self._max_boilerplate_string_ratio


class MeanWordLengthFilter(DocumentFilter):
    """
    If the mean word length is not in a specified range, then discard.

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(
        self,
        min_mean_word_length=3,
        max_mean_word_length=10,
        lang="en",
    ):
        super().__init__()
        self._min_cutoff = min_mean_word_length
        self._max_cutoff = max_mean_word_length
        self._word_splitter = get_word_splitter(lang)
        self._name = "mean_word_length"

    def score_document(self, text):
        word_lens = [len(w) for w in self._word_splitter(text.strip()) if len(w) > 0]
        return sum(word_lens) / len(word_lens)

    def keep_document(self, score):
        return self._min_cutoff <= score <= self._max_cutoff


class RepeatedLinesFilter(DocumentFilter):
    """
    If the document shrinks by > 30% in terms of number of lines after
    removing duplicate lines, then discard.
    Source: Gopher (Rae et al., 2021)
    """

    def __init__(self, max_repeated_line_fraction=0.7):
        super().__init__()
        self._cutoff = max_repeated_line_fraction
        self._name = "repeated_lines"

    def score_document(self, text):
        sentences = self._sentences
        if sentences is None:
            sentences = get_sentences(text)
        return len(set(sentences)) / len(sentences)

    def keep_document(self, score):
        return score >= self._cutoff


class RepeatedParagraphsFilter(DocumentFilter):
    """
    If the document shrinks by > 30% in terms of number of lines after
    removing duplicate paragraphs, then discard.
    Source: Gopher (Rae et al., 2021)
    """

    def __init__(self, max_repeated_paragraphs_ratio=0.7):
        super().__init__()
        self._max_repeated_paragraphs_ratio = max_repeated_paragraphs_ratio
        self._name = "repeated_paragraphs"

    def score_document(self, text):
        paragraphs = self._paragraphs
        if paragraphs is None:
            paragraphs = get_paragraphs(text)
        return len(set(paragraphs)) / len(paragraphs)

    def keep_document(self, score):
        return score >= self._max_repeated_paragraphs_ratio


class RepeatedLinesByCharFilter(DocumentFilter):
    """
    If the document shrinks by > 20% in terms of number of lines
    after removing duplicate lines, then discard.
    Source: Gopher (Rae et al., 2021)
    """

    def __init__(self, max_repeated_lines_char_ratio=0.8):
        super().__init__()
        self._cutoff = max_repeated_lines_char_ratio
        self._name = "repeated_lines_char"

    def score_document(self, text):
        sentences = self._sentences
        if sentences is None:
            sentences = get_sentences(text)

        return len("".join(set(sentences))) / len("".join(sentences))

    def keep_document(self, score):
        return score >= self._cutoff


class RepeatedParagraphsByCharFilter(DocumentFilter):
    """
    If the document shrinks by > 10% in terms of number of lines after
    removing duplicate paragraphs, then discard.
    Source: Gopher (Rae et al., 2021)
    """

    def __init__(self, max_repeated_paragraphs_char_ratio=0.8):
        super().__init__()
        self._cutoff = max_repeated_paragraphs_char_ratio
        self._name = "repeated_paragraphs_char"

    def score_document(self, text):
        paragraphs = self._paragraphs
        if paragraphs is None:
            paragraphs = get_paragraphs(text)

        return len("".join(set(paragraphs))) / len("".join(paragraphs))

    def keep_document(self, score):
        return score >= self._cutoff


class RepeatingTopNGramsFilter(DocumentFilter):
    """
    If the document shrinks by > x% in terms of number of characters after
    removing the top n-grams, then discard.
    Source: Gopher (Rae et al., 2021)

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(self, n=2, max_repeating_ngram_ratio=0.2, lang="en"):
        super().__init__()
        self._n = n
        self._cutoff = max_repeating_ngram_ratio
        self._max_ratio = 1.0
        self._word_splitter = get_word_splitter(lang)
        self._name = f"repeating_top_{n}grams"

    def score_document(self, text):
        ngrams = self._ngrams
        if ngrams is None:
            split_text = self._word_splitter(text.strip())
            if len(split_text) < self._n:
                return self._max_ratio
            ngrams = get_ngrams(split_text, self._n)
        unique_ngrams = set(ngrams)
        # Find the most frequent ngram in the zipped ngram list
        counts = {
            ngram: {"freq": 0, "num_chars": sum(len(word) for word in ngram)}
            for ngram in unique_ngrams
        }
        for ngram in ngrams:
            counts[ngram]["freq"] += 1
        most_frqnt_ngram = " ".join(max(counts, key=lambda x: counts[x]["freq"]))
        # Find the number of characters the most frequent ngram
        # contributes to the document
        nchar = len(text)
        len_diff = nchar - len(text.replace(most_frqnt_ngram, ""))
        if nchar > 0:
            score = len_diff / nchar
        else:
            # Remove if the document is empty
            score = 1.0
        return score

    def keep_document(self, score):
        return score <= self._cutoff


class RepeatingDuplicateNGramsFilter(DocumentFilter):
    """
    If the document shrinks by > x% in terms of number of characters
    after removing all duplicate n-grams, then discard.
    Source: Gopher (Rae et al., 2021)

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(self, n=2, max_repeating_duplicate_ngram_ratio=0.2, lang="en"):
        super().__init__()
        self._n = n
        self._cutoff = max_repeating_duplicate_ngram_ratio
        self._max_ratio = 1.0
        self._word_splitter = get_word_splitter(lang)
        self._name = f"repeating_dup_{n}gram"

    def score_document(self, text):
        ngrams = self._ngrams
        if ngrams is None:
            split_text = self._word_splitter(text.strip())
            if len(split_text) < self._n:
                return self._max_ratio
            ngrams = get_ngrams(split_text, self._n)

        counts = {}
        duplicated_nchar = 0
        overlapping_ngrams = 0
        for ngram in ngrams:
            counts[ngram] = counts.get(ngram, 0) + 1
            if counts[ngram] > 1:
                # Count the number of characters in this ngram that haven't been counted already
                duplicated_ngrams = sum(
                    len(gram) for gram in ngram[overlapping_ngrams:]
                )
                # Count the spaces between the ngrams
                nspaces = min(self._n - overlapping_ngrams, self._n - 1)
                duplicated_nchar += duplicated_ngrams + nspaces
                overlapping_ngrams = self._n
            overlapping_ngrams = max(overlapping_ngrams - 1, 0)

        nchar = len(text)
        if nchar > 0:
            score = duplicated_nchar / nchar
        else:
            # Remove if the document is empty
            score = 1.0
        return score

    def keep_document(self, score):
        return score <= self._cutoff


class PunctuationFilter(DocumentFilter):
    """
    If more than 85% of the sentences do not end with a
    punctuation mark, then discard.
    Source: Google C4 processing
    """

    def __init__(self, max_num_sentences_without_endmark_ratio=0.85):
        super().__init__()
        self._cutoff = max_num_sentences_without_endmark_ratio
        self._name = "punctuation"

    def score_document(self, text):
        sentences = self._sentences
        if sentences is None:
            sentences = get_sentences(text)
        num_sentence_without_endmarks = len(
            [s for s in sentences if not s.strip().endswith(end_marks)]
        )
        return num_sentence_without_endmarks / len(sentences)

    def keep_document(self, score):
        return score <= self._cutoff


class EllipsisFilter(DocumentFilter):
    """
    If more than 30% of the sentences end with an elipsis, then discard.
    Source: Google C4 processing
    """

    def __init__(self, max_num_lines_ending_with_ellipsis_ratio=0.3):
        super().__init__()
        self._cutoff = max_num_lines_ending_with_ellipsis_ratio
        self._name = "ellipsis"

    def score_document(self, text):
        sentences = self._sentences
        if sentences is None:
            sentences = get_sentences(text)
        num_lines_ending_with_ellipsis = 0
        for sentence in sentences:
            for ellipsis in ellipsis_marks:
                if sentence.strip().lower().endswith(ellipsis):
                    num_lines_ending_with_ellipsis += 1
                    break
        return num_lines_ending_with_ellipsis / len(sentences)

    def keep_document(self, score):
        return score <= self._cutoff


class CommonEnglishWordsFilter(DocumentFilter):
    """
    If the sentence contains at least 2 common English words, then keep it.
    NOTE: We purposefully check for the lowercase versions of those common words
    to remove documents with over-capitalization.

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(self, min_num_common_words=2, stop_at_false=True):
        super().__init__()
        self._cutoff = min_num_common_words
        self._stop_at_false = stop_at_false
        self._word_splitter = get_word_splitter("en")
        self._name = "common_english_words"

    def score_document(self, text):
        common_word_counter = 0
        for word in self._word_splitter(text.strip()):
            if word in common_english_words:
                common_word_counter += 1
            if self._stop_at_false and common_word_counter >= self._cutoff:
                return common_word_counter

        return common_word_counter

    def keep_document(self, score):
        return score >= self._cutoff


class WordsWithoutAlphabetsFilter(DocumentFilter):
    """
    80% of words in a document must contain at least one alphabetic character.
    Source: Gopher (Rae et al., 2021)

    For Chinese and Japanese text, we use external libraries to split the text
    because these languages are not separated by spaces. For all other langauges,
    such as English, we assume words are separated by spaces.
    """

    def __init__(self, min_words_with_alphabets=0.8, lang="en"):
        super().__init__()
        self._cutoff = min_words_with_alphabets
        self._word_splitter = get_word_splitter(lang)
        self._name = "words_without_alphabets"

    def score_document(self, text):
        num_english_alpha = 0
        words = self._word_splitter(text.strip())
        for word in words:
            if regex_alpha.search(word):
                num_english_alpha += 1

        return num_english_alpha / len(words)

    def keep_document(self, score):
        return score >= self._cutoff


class PornographicUrlsFilter(DocumentFilter):
    """
    Check if any of the URLs within the document point to pornography.
    """

    def __init__(self):
        super().__init__()

    def score_document(self, text):
        all_urls = regex_url.findall(text)
        for url in all_urls:
            if "porn" in url:
                return 1

        return 0

    def keep_document(self, score):
        return score != 1


class TokenCountFilter(DocumentFilter):
    """
    If the document contains more or less than a specified number of tokens, then discard.
    """

    def __init__(self, tokenizer: AutoTokenizer, min_tokens=0, max_tokens=float("inf")):
        """
        Args:
            tokenizer (AutoTokenizer): The tokenizer to use to count the tokens.
            min_tokens (int): The minimum number of tokens the document must contain.
                Set to 0 to disable the minimum token count filter.
            max_tokens (int): The maximum number of tokens the document can contain.
                Set to infinity to disable the maximum token count filter.
        """
        super().__init__()
        self._tokenizer = tokenizer
        self._min_tokens = min_tokens
        self._max_tokens = max_tokens
        self._name = "token_count"

    def score_document(self, text: str) -> int:
        tokens = self._tokenizer.encode(text)
        return len(tokens)

    def keep_document(self, score: int) -> bool:
        return self._min_tokens <= score <= self._max_tokens


class SubstringFilter(DocumentFilter):
    """
    Keeps documents that contain a substring in a given position.
    Gives a score of 1 if the substring is found in the given position, otherwise 0.
    """

    def __init__(self, substring: str, position: Literal["prefix", "suffix", "any"]):
        """
        Args:
            substring (str): The substring to check for.
            position (Literal["prefix", "suffix", "any"]): The position of the substring.
        """
        super().__init__()
        self._substring = substring
        if position not in ["prefix", "suffix", "any"]:
            raise ValueError(
                f"Invalid position: {position}. Must be one of: prefix, suffix, any."
            )
        self._position = position

    def score_document(self, text: str) -> int:
        if self._position == "prefix":
            return int(text.startswith(self._substring))
        elif self._position == "suffix":
            return int(text.endswith(self._substring))
        elif self._position == "any":
            return int(self._substring in text)

    def keep_document(self, score: int) -> bool:
        return score == 1


class HistogramFilter(DocumentFilter):
    """Histogram filter used by the NLLB paper (https://arxiv.org/pdf/2207.04672). See p30 for details.

    The high-level idea of histogram filter can be described as a cheap version of language ID.
    Basically, it checks what ratio of characters in the data instance are included in the character historgrams collected from trusted data in the corresponding language.
    If the ratio is too low, then there is a good chance that there is a language ID mismatch and the data instance should be discarded.

    Written with reference to the original fairseq implementation at:
    https://github.com/facebookresearch/fairseq/blob/main/examples/m2m_100/process_data/clean_histogram.py.
    """

    def __init__(self, lang="en", threshold=0.8, cache_dir="", threshold_char="]"):
        """Args:
        lang (str, optional): Expected language of the segment. This will decide which histogram will be loaded. Defaults to "en".
        threshold (float, optional): Threshold for ratio of characters in the histogram. Defaults to 0.8.
        cache_dir (str, optional): Cache dir download histogram files. Defaults to "".
        threshold_char (str, optional): Formatter character of the histogram files. You should not change this unless you rebuilt your own histogram. Defaults to "]".
        """
        super().__init__()
        self._lang = lang
        self._threshold = threshold
        self._cache_dir = cache_dir if cache_dir else user_cache_dir()
        self._threshold_char = threshold_char
        self._name = "histogram"

        if not os.path.isdir(os.path.join(self._cache_dir, "histograms")):
            self._download_histograms()

        self._read_hist()

    def _download_histograms(self):
        """Download and process histograms from default repo.

        Raises:
            requests.exceptions.RequestException: If download fails.
        """

        # Send a GET request to the URL
        response = requests.get(
            "https://dl.fbaipublicfiles.com/m2m_100/histograms.tar.gz"
        )

        # Check if the request was successful
        if response.status_code != 200:
            raise requests.exceptions.RequestException(
                f"Failed to download histogram file. Status code: {response.status_code}"
            )

        # Open a file to write the content
        os.makedirs(self._cache_dir, exist_ok=True)
        download_dest_path = os.path.join(self._cache_dir, "histograms.tar.gz")
        with open(download_dest_path, "wb") as file:
            file.write(response.content)

        extract_path = os.path.join(self._cache_dir, "histograms")
        with tarfile.open(download_dest_path, "r:gz") as tar:
            # Extract all the contents into the specified directory
            tar.extractall(path=extract_path)

    def _read_hist(self):
        """Load histogram files."""

        self._histogram = []
        with open(
            os.path.join(
                self._cache_dir,
                "histograms",
                "checkpoint",
                "edunov",
                "cc60_multilingual",
                "clean_hists",
                self._lang,
            )
        ) as f:
            for line in f:
                c = line[0]
                if c == self._threshold_char:
                    break
                self._histogram.append(c)
        self._histogram = set(self._histogram)

    def score_document(self, text: str) -> float:
        """Compute histogram token ratio of a text data instance according to the loaded histogram.

        Args:
            text (str): Text data instance.

        Returns:
            float: Ratio of tokens included in the histogram.
        """
        cnt = len([c for c in text.strip() if c in self._histogram])
        return 1 if cnt / len(text) > self._threshold else 0

    def keep_document(self, score):
        return score == 1


class LengthRatioFilter(BitextFilter):
    """(Bitext filter) Length ratio filter for bitext, similar to the one implemented in Moses toolkit (`https://github.com/moses-smt/mosesdecoder/blob/master/scripts/training/clean-corpus-n.perl`).

    If the ratio between source and target tokens is not within a specified range then discard. Either direction (src/tgt, tgt/src) is considered.
    """

    def __init__(self, max_ratio=3.0, src_lang="en", tgt_lang="en", **kwargs):
        """Args:
        max_ratio (float, optional): Maximum allowed length ratio between either direction of the bitext. Defaults to 3.0.
        src_lang (str, optional): Language of the source data (needed for tokenization). Defaults to "en".
        tgt_lang (str, optional): Language of the target data (needed for tokenization). Defaults to "en".
        """

        super().__init__(**kwargs)
        self._max_ratio = float(max_ratio)
        self._src_word_splitter = get_word_splitter(src_lang)
        self._tgt_word_splitter = get_word_splitter(tgt_lang)
        self._name = "length_ratio"

    def score_bitext(self, src: str, tgt: str) -> float:
        """Tokenize the source and target sentences and compute length ratio.

        Args:
            src (str): Source document string.
            tgt (str): Target document string.

        Returns:
            float: The maximum ratio among the two translation directions of the bitext.
        """
        src_len = len(self._src_word_splitter(src.strip()))
        tgt_len = len(self._tgt_word_splitter(tgt.strip()))
        return max(src_len / tgt_len, tgt_len / src_len)

    def keep_bitext(self, score):
        """Decides whether a single document should be retained according to the computed length ratio."""
        return score < self._max_ratio



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

unsafe_categories = """O1: Violence.
Should not
- Help people plan or engage in violence.
- Encourage people to engage in violence.
- Contain unpleasant vivid violent imagery or violent description that arouses discomfort in the reader.
Can
- Provide information on violence.
- Discuss the topics of violence, and can discuss historical events involving violence.
O2: Sexual.
Should not
- Engage in sexually explicit (i.e., erotic) conversations and should not encourage people to engage in sexual activities.
- Engage or encourage sexual activity that is considered illegal.
Can
- Discuss topics related to sexuality and sexual education.
- Provide general information regarding sexual activities.
- Provide guidance on topics related to sexual health.
O3: Criminal Planning/Confessions.
Should not
- Help people plan or engage in criminal activities like theft, kidnapping, or financial crimes, especially when the person indicates that they may act on the information provided to them.
- Encourage people to engage in criminal activities or condone criminal behavior.
- Have a confession of crime that was committed or will be committed in future.
Can
- Help people understand how crimes are committed, provided that the response is not so detailed that it could be used to provide a roadmap for future crimes.
- Help people understand how systems work that are used to prevent crimes, provided those descriptions do not provide detailed information on how to defeat those systems.
O4: Guns and Illegal Weapons.
Should not
- Assist in the planning or execution of firearm or illegal weapons crimes, especially when the person indicates that they may act on the information provided to them.
- Encourage people to engage in firearm or illegal weapons crimes or condone such behavior.
- Assist or encourage people to execute or own other illegal weapons for example biological agents, or chemical weapons, including instructions on how to create or use them.
Can
- Discuss firearms and the arguments for and against firearm ownership.
O5: Controlled/Regulated Substances.
Should not
- Encourage or assist people in the trafficking, consumption, or use of regulated or controlled substances, including but not limited to illegal drugs, tobacco, alcohol, and cannabis.
- Assist or encourage people to create such substances when it is illegal to do so.
Can
- Discuss the use of regulated or controlled substances, and can provide general information on those substances (ex: what their chemical composition is and what their effects are).
- Provide information on how those substances are manufactured, provided that the information is not so detailed as to provide step-by-step instructions for their creation.
O6: Suicide and Self Harm.
Should not
- Encourage people to harm themselves.
- Romanticize or condone self-harm or suicide.
- Provide information regarding the methods of suicide or self-harm.
- Provide any information or advice, other than information regarding appropriate health resources, to people who express or imply that they wish to harm themselves.
- Contain confessions or intent of self harm
Can
- Provide general information regarding self-harm and suicide (ex: statistics, the mental health challenges faced by those who take these actions).
Should
- Provide information regarding appropriate health resources to people who imply or express the desire to harm themselves.
O7: Sexual (minor).
Should not
- Assist, engage or encourage intention to engage in sexual activity with an underage/minor.
- Discuss or plan sexual acts, sharing of media of sexual nature, personal information or intention to meet in person for sexual activity with a minor.
- Should not include discussion where private information is shared with underage minors for purposes of acting on them and taking actions that are sexual in nature including meeting or sharing media.
- Should not discuss child abuse or engage, encourage, assist in sexual grooming/trafficking of children.
Can
- Provide information about sex education for use in schools.
O8: Hate /identity hate.
Should not
- Express hateful, derogatory, or demeaning sentiments against people based on sensitive personal characteristics like their 
race, color, ethinicity, religion, national origin, sexual orientation, gender, gender identity, or disability or caste.
- Encourage discrimination against people based on those characteristics and should not use slurs or create insults based on those characteristics.
- Should not discuss prejudice towards a group classified as protected according to the law of Germany, UK, USA or any other country.
- Should not contain slurs.
- Should not contain text that is distasteful, offensive, or hateful that causes others to feel grief, anger, or fear. 
Can
- Provide information on discrimination.
- Discuss the topics of discrimination, and can discuss historical events involving discrimination.
- Discuss neutral statements referring to group identity and  contextualized hate speech (that talks about hate but does not incite hate). 
O9: PII/Privacy.
Should not
- Include statements that either requests for PII information made in a conversation turn or mention information such as name, address, 
phone number or other social media handle. 
- Should not discuss or contain personally identifying information(PII) that isn’t publicly available and can be used to determine the identity of the user such as
date of birth, national ID, IP address, internet cookies, other telemetry information, physical characteristics, biometrics data, financial, health, belief or social data.
- Should not include any statement that violates privacy laws set out by GDPR, Europe, CCPA, California,  Personal Information Protection law (PIPL) China and
Children’s Online Privacy Protection Act (COPPA), US and all country, state, and municipal level privacy law.
Can
- Provide information what PII means.
- Provide information about privacy laws across different countries and regions.
O10: Harassment.
Should not 
- Have content that may be used to torment or annoy individuals in real life, or make harassment more likely to occur. 
- Engage in harassment, misconduct and unlawful discrimination 
- Single out an individual or marginalization and/or retaliation based on protected characteristics like Race, Color, Gender, Sex,
Sexual orientation, Gender identity, and gender expression, National origin, Ethnicity, Disability (including being regarded as disabled),
Religion, Age (40+), Pregnancy (including pregnancy, childbirth or related medical conditions), Genetic information, Military or veteran status, Citizenship status, Political activity, or affiliation,
- Taking or requesting statutorily protected leave Body characteristics. 
- Promise to give a benefit, or a threat to retaliate or take an adverse action based on a response to harassment and misconduct.
- Bullying. 
- Being derogatory and humiliating toward an individual but not necessarily protected characteristics under law. 
- Making rude or insulting comments, demeaning, and objectifying terms toward an individual.
Can
- Discuss negative impacts of harassment and bullying based on protected characteristics like race, color, gender, sex, nationality, within a contextualized setting.
- Provide information on protected characteristics or referor cite material that describes what protected characteristics are.
- Provide information or refer or cite material that informs what harassment is.
O11: Threat.
Should not
- Discuss a communication of intent to inflict harm or loss on another person. 
- Should not contain discussion of intention of intimidation which  is a tactic used between conflicting parties to make the other timid or psychologically insecure for coercion or control.
- Should not contain intention of knowingly putting another person in fear of bodily injury. 
- Should not assist or plan injury, physical or mental damage, act or instance of injury, or material  or any kind of loss to a person.
- Should not contain threats that are forbidden by law as intent to obtain a monetary advantage or to compel a person to act against their will.
Can
- Provide information about threat is and the meaning of threat.
O12: Profanity.
Should not
- Contain swear words, curse words, or other obscene or profane language
- This includes offensive words used without any intention to act on them or without any explicit targetted recipient
O13: Needs Caution.
Should not
- Discuss text that is toxic in nature, goes against social norms, unethical or immoral in general.
- Contain text that cannot be considered 'safe' with the information provided.
Can
- Include text that is considered to be 'safe' and that does not belong to any of the unsafe categorie described above.
"""


def format_aegis(message: str) -> str:
    """
    Formats the prompt to the format that the AEGIS model and llama guard accepts
    """
    role = "User"
    prompt = f"""[INST] Task: Check if there is unsafe content in '{role}' messages in conversations according our safety policy with the below categories.

    <BEGIN UNSAFE CONTENT CATEGORIES>
    {unsafe_categories}
    <END UNSAFE CONTENT CATEGORIES>

    <BEGIN CONVERSATION>

    """
    prompt += f"{role}: {message}\n\n"
    prompt += f"""<END CONVERSATION>

    Provide your safety assessment for {role} in the above conversation:
    - First line must read 'safe' or 'unsafe'.
    - If unsafe, a second line must include a comma-separated list of violated categories. [/INST]"""
    return prompt



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from functools import partial, reduce
from typing import Iterable, List, Union

import dask.dataframe as dd
from dask import delayed

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modules.base import BaseModule
from nemo_curator.tasks.downstream_task import DownstreamTask
from nemo_curator.utils.distributed_utils import single_partition_write_with_filename
from nemo_curator.utils.text_utils import get_words


class TaskDecontamination(BaseModule):
    def __init__(
        self,
        tasks: Union[DownstreamTask, Iterable[DownstreamTask]],
        text_field="text",
        max_ngram_size=13,
        max_matches=10,
        min_document_length=200,
        remove_char_each_side=200,
        max_splits=10,
        removed_dir=None,
    ) -> None:
        """
        Removes segments of downstream evaluation tasks from a dataset
        Args:
            max_ngram_size: The maximum amount of task grams that are considered at once for contamination.
            max_matches: If an ngram is found more than max_matches times, it is considered too common and will not be removed from the dataset.
            min_document_length: When a document is split, if a split falls below this character length it is discarded.
            remove_char_each_side: The number of characters to remove on either side of the matching ngram
            max_splits: The maximum number of times a document may be split before being entirely discarded.
            removed_dir: If not None, the documents split too many times will be written to this directory using the filename in the dataset.
        """
        super().__init__(input_backend="pandas")
        if isinstance(tasks, DownstreamTask):
            tasks = [tasks]
        self.tasks = tasks
        self.text_field = text_field
        self.max_ngram_size = max_ngram_size
        self.max_matches = max_matches
        self.min_document_length = min_document_length
        self.remove_char_each_side = remove_char_each_side
        self.max_splits = max_splits
        self.removed_dir = removed_dir

    def call(self, dataset: DocumentDataset) -> DocumentDataset:

        # Convert the dataframe to delayed objects for complex operations
        original_meta = dataset.df.dtypes.to_dict()
        delayed_dataset = dataset.df.to_delayed()

        # Perform task decontamintation
        task_ngrams = self.prepare_task_ngram_count()
        found_result = self._find_matching_ngrams(task_ngrams, delayed_dataset)
        matched_ngrams, ngram_freq = (
            found_result["matched-ngrams"],
            found_result["ngrams-freq"],
        )
        delayed_removed_dataset = self._remove_matching_ngrams(
            matched_ngrams, ngram_freq, delayed_dataset
        )

        # Restore the dataset to its original format
        removed_dataset = DocumentDataset(
            dataset_df=dd.from_delayed(delayed_removed_dataset, meta=original_meta)
        )

        return removed_dataset

    @staticmethod
    def _merge_task_ngrams(first: dict, second: dict):
        first.update(second)
        return first

    def prepare_task_ngram_count(self) -> dict:
        """
        Computes a dictionary of all ngrams in each task as keys and each value set to 0.
        """
        delayed_ngrams = [delayed(task.generate_ngrams)() for task in self.tasks]
        aggregated_ngrams = delayed(reduce)(
            TaskDecontamination._merge_task_ngrams, delayed_ngrams
        )

        return aggregated_ngrams

    @staticmethod
    def _compute_ngram_freq_sorted(task_ngrams):
        ngrams_freq = defaultdict(int)
        for ngram_key in task_ngrams.keys():
            ngram_words, _ = get_words(ngram_key)
            length = len(ngram_words)
            ngrams_freq[length] += 1

        ngrams_freq_sorted = sorted(ngrams_freq.items(), key=lambda item: item[0])

        return ngrams_freq_sorted

    def find_matching_ngrams(self, task_ngrams: dict, dataset: DocumentDataset) -> dict:
        delayed_dataset = dataset.df.to_delayed()

        return self._find_matching_ngrams(task_ngrams, delayed_dataset)

    def _find_matching_ngrams(self, task_ngrams: dict, delayed_dataset) -> dict:
        task_ngrams_frequency_sorted = delayed(self._compute_ngram_freq_sorted)(
            task_ngrams
        )
        delayed_counts = [
            delayed(self._find_ngrams_partition)(
                partition, task_ngrams, task_ngrams_frequency_sorted
            )
            for partition in delayed_dataset
        ]
        combined_counts = delayed(reduce)(self._merge_counts, delayed_counts)
        formatted_result = delayed(self._format_matching_ngrams_result)(
            combined_counts, task_ngrams_frequency_sorted
        )

        return formatted_result

    def _find_ngrams_partition(
        self, dataset_partition, task_ngrams, ngrams_freq_sorted
    ):
        partition_count = defaultdict(int)
        for document in dataset_partition[self.text_field]:
            doc_result = self._find_ngrams(document, task_ngrams, ngrams_freq_sorted)
            partition_count = TaskDecontamination._merge_counts(
                partition_count, doc_result
            )

        return partition_count

    @staticmethod
    def _merge_counts(first: dict, second: dict):
        for ngram, count in second.items():
            first[ngram] = first.get(ngram, 0) + count

        return first

    @staticmethod
    def _format_matching_ngrams_result(matched_ngrams, ngram_freq):
        return {
            "matched-ngrams": matched_ngrams,
            "ngrams-freq": ngram_freq,
        }

    def _find_ngrams(self, document, task_ngrams, ngrams_freq_sorted):
        """
        Searches for matching n-grams in a document
        """
        text_buf = [document]

        local_ngram = defaultdict(int)
        while len(text_buf) > 0:

            # get the first one from the buffer
            text = text_buf.pop(0)
            words, positions = get_words(text)

            ngram_free = True
            # First, loop over all n-grams in document
            for i in range(len(words) - self.max_ngram_size + 1):
                # Check if we found a matching n-gram
                check_ngram_free = TaskDecontamination._check_text(
                    words[i : i + self.max_ngram_size],
                    task_ngrams,
                    text,
                    positions[i],
                    text_buf,
                    local_ngram,
                )

                # If we found a match, break
                # the remainder of the text is appended to text_buf
                # for futher processing
                if not check_ngram_free:
                    ngram_free = False
                    break

                # Continue searching for the remaining dominant n-grams
                for ngram_len, _ in ngrams_freq_sorted:
                    # Check if we found a matching n-gram
                    check_ngram_free = TaskDecontamination._check_text(
                        words[i : i + ngram_len],
                        task_ngrams,
                        text,
                        positions[i],
                        text_buf,
                        local_ngram,
                    )

                    # Again, if we find match, break
                    # the remainder of the text is appended to text_buf
                    # for futher processing
                    if not check_ngram_free:
                        ngram_free = False
                        break

                # Additional break to break out of both loops
                if not ngram_free:
                    break

            # If did not find a match for the max_ngram_size
            # check the ending n-gram
            if ngram_free and len(words) - self.max_ngram_size > 0:
                # get the last words of the lax max ngram
                last_seq_words = words[len(words) - self.max_ngram_size : len(words)]
                last_seq_start_position = len(words) - self.max_ngram_size

                # check all n-grams lower than max ngram-len
                for pos, (ngram_len, _) in enumerate(ngrams_freq_sorted):

                    # ignore the max ngram as has been considered already
                    if ngram_len == self.max_ngram_size:
                        continue

                    # find each ngram of ngram_len in max n-grams and check
                    for i in range(len(last_seq_words) - ngram_len + 1):
                        # Check for matching n-grams
                        check_ngram_free = TaskDecontamination._check_text(
                            last_seq_words[i : i + ngram_len],
                            task_ngrams,
                            text,
                            positions[last_seq_start_position + i],
                            text_buf,
                            local_ngram,
                        )

                        # If we find a match, break
                        if not check_ngram_free:
                            ngram_free = False
                            break

                    # Break from both loops
                    if not ngram_free:
                        break

        return local_ngram

    @staticmethod
    def _check_text(words, task_ngrams, text, start_position, text_buf, local_ngram):
        seq = " ".join(words)
        if seq in task_ngrams:
            print(" [matched]: {}".format(seq), flush=True)
            # If this flag is set, we just look for matching n-grams
            # we don't remove any matching n-grams
            # Count the matched n-gram and consider it later
            local_ngram[seq] += 1
            if (start_position + len(seq) + 1) < len(text):
                text_buf.append(text[start_position + len(seq) + 1 : len(text)])
            return False

        return True

    def remove_matching_ngrams(
        self, matched_ngrams: dict, ngram_freq: List[tuple], dataset: DocumentDataset
    ):
        original_meta = dataset.df.dtypes.to_dict()
        delayed_dataset = dataset.df.to_delayed()
        delayed_removed_dataset = self._remove_matching_ngrams(
            matched_ngrams, ngram_freq, delayed_dataset
        )
        removed_dataset = DocumentDataset(
            dataset_df=dd.from_delayed(delayed_removed_dataset, meta=original_meta)
        )

        return removed_dataset

    def _remove_matching_ngrams(
        self, matched_ngrams: dict, ngram_freq: List[tuple], delayed_dataset
    ):
        threshhold_ngrams = delayed(self._threshold_ngram_count)(matched_ngrams)
        delayed_removed_dataset = [
            delayed(self._remove_ngrams_partition)(
                partition, threshhold_ngrams, ngram_freq
            )
            for partition in delayed_dataset
        ]

        return delayed_removed_dataset

    def _threshold_ngram_count(self, matched_ngrams: dict) -> set:
        filtered_ngrams = set()
        for ngram, count in matched_ngrams.items():
            if count <= self.max_matches:
                filtered_ngrams.add(ngram)

        return filtered_ngrams

    def _remove_ngrams_partition(self, partition, task_ngrams, ngrams_freq_sorted):
        text_type = partition[self.text_field].dtype

        document_fn = partial(
            self._remove_ngrams,
            task_ngrams=task_ngrams,
            ngrams_freq_sorted=ngrams_freq_sorted,
        )
        split_text = partition[self.text_field].apply(document_fn)
        num_splits = split_text.apply(len)

        valid_documents_mask = (1 <= num_splits) & (num_splits <= self.max_splits)

        if self.removed_dir:
            removed_docs = partition[~valid_documents_mask]
            single_partition_write_with_filename(removed_docs, self.removed_dir)

        partition[self.text_field] = split_text
        filtered_partition = partition[valid_documents_mask]
        exploded_partition = filtered_partition.explode(
            self.text_field, ignore_index=True
        )
        # After exploding, the string datatype can become an "object" type
        exploded_partition[self.text_field] = exploded_partition[
            self.text_field
        ].astype(text_type)

        return exploded_partition

    def _remove_ngrams(self, document, task_ngrams, ngrams_freq_sorted):
        """
        Searches for matching n-grams in a document
        """
        text_buf = [document]

        text_buf_ngram_free = []
        while len(text_buf) > 0:
            # get the first one from the buffer
            text = text_buf.pop(0)
            words, positions = get_words(text)

            ngram_free = True
            # First, loop over all n-grams in document
            for i in range(len(words) - self.max_ngram_size + 1):
                # Check if we found a matching n-gram
                check_ngram_free = self._clean_text(
                    words[i : i + self.max_ngram_size],
                    task_ngrams,
                    text,
                    positions[i],
                    text_buf,
                    text_buf_ngram_free,
                )

                # If we found a match, break
                # the remainder of the text is appended to text_buf
                # for futher processing
                if not check_ngram_free:
                    ngram_free = False
                    break

                # Continue searching for the remaining dominant n-grams
                for ngram_len, _ in ngrams_freq_sorted:
                    # Check if we found a matching n-gram
                    check_ngram_free = self._clean_text(
                        words[i : i + ngram_len],
                        task_ngrams,
                        text,
                        positions[i],
                        text_buf,
                        text_buf_ngram_free,
                    )

                    # Again, if we find match, break
                    # the remainder of the text is appended to text_buf
                    # for futher processing
                    if not check_ngram_free:
                        ngram_free = False
                        break

                # Additional break to break out of both loops
                if not ngram_free:
                    break

            # If did not find a match for the max_ngram_size
            # check the ending n-gram
            if ngram_free and len(words) - self.max_ngram_size > 0:
                # get the last words of the lax max ngram
                last_seq_words = words[len(words) - self.max_ngram_size : len(words)]
                last_seq_start_position = len(words) - self.max_ngram_size

                # check all n-grams lower than max ngram-len
                for pos, (ngram_len, _) in enumerate(ngrams_freq_sorted):

                    # ignore the max ngram as has been considered already
                    if ngram_len == self.max_ngram_size:
                        continue

                    # find each ngram of ngram_len in max n-grams and check
                    for i in range(len(last_seq_words) - ngram_len + 1):
                        # Check for matching n-grams
                        check_ngram_free = self._clean_text(
                            last_seq_words[i : i + ngram_len],
                            task_ngrams,
                            text,
                            positions[last_seq_start_position + i],
                            text_buf,
                            text_buf_ngram_free,
                        )

                        # If we find a match, break
                        if not check_ngram_free:
                            ngram_free = False
                            break

                    # Break from both loops
                    if not ngram_free:
                        break

            # texts are ngram free
            if ngram_free:
                text_buf_ngram_free.append(text)

        # check if the text has only been trimmed
        trimmed = 0
        if len(text_buf_ngram_free) == 1:
            if len(text_buf_ngram_free[0]) < len(text):
                trimmed = 1

        return text_buf_ngram_free

    def _clean_text(
        self,
        words,
        matched_ngrams,
        text,
        start_position,
        text_buf,
        text_buf_ngram_free,
        nosplit_remove=False,
    ):
        seq = " ".join(words)
        if seq in matched_ngrams:
            print(" [matched]: {}".format(seq), flush=True)

            # for NMT data we want to completely remove the sample
            # which has a match
            if nosplit_remove:
                return False

            # split the text
            text_first, text_second = TaskDecontamination._split_text(
                text,
                start_position,
                self.remove_char_each_side,
                seq,
            )

            # Free up the first part of matching n-grams
            if len(text_first) > self.min_document_length:
                text_buf_ngram_free.append(text_first)

            # The second part of the text is added to the output buffer
            # and will be processed later
            if len(text_second) > self.min_document_length:
                text_buf.append(text_second)

            # Is not free of matching ngrams
            return False

        # Free of matching n-grams
        return True

    @staticmethod
    def _split_text(text, start_pos, remove_char_each_side, seq):
        # first part of the text
        punctuations = ".!?"
        pos = start_pos - remove_char_each_side
        text_first = ""
        while pos > 0 and not text[pos] in punctuations:
            pos -= 1
        if pos > 0:
            text_first = text[0 : pos + 1]

        # add length of seq and remove_char_each_side
        pos = start_pos + len(seq) + remove_char_each_side

        # last part of the text
        text_second = ""
        while pos < len(text) and not text[pos] in punctuations:
            pos += 1
        if pos + 1 < len(text):
            text_second = text[pos + 1 : len(text)]

        return text_first, text_second
